import numpy as np
from mindaffectBCI.decoder.decodingSupervised import decodingSupervised
def decodingCurveSupervised(Fy,objIDs=None,nInt=(30,25),**kwargs):
    '''
    Compute a decoding curve, i.e. mistake-probability over time for probability based stopping from the per-epoch output scores
    
    Args:
        Fy  = (nModel,nTrl,nEp,nY) [nY x nEpoch x nTrl x nModel ] similarity score for each input epoch for each output
                N.B. Supervised decoder only has 1 model!!!
        objIDs = (nY,) mapping from rows of Fy to output object IDs.  N.B. assumed objID==0 is true target
                N.B. if objIDs > size(Fy,2), then additional virtual outputs are added
        nInt = (2,) the number of integeration lengths to use, numThresholds. Defaults to ([30,25])
    Returns:
        integerationLengths - [nInt] the actual integeration lengths in samples
        ProbErr: (nInt,)float empherical error probablility at this integeration length
        ProbErrEst(np.ndarray): (nInt: float) decoder estimate of the error rate for each integeration length
        StopPerr - (nInt,) error rate at this average trial length when using ProbErrEst-thresholding based stopping
        StopThresh - (nInt,) ProbErrEst threshold used to get this average trial length.
        Yerr(np.ndarray): (nTrl,nInt : bool) flag if estimate was correct at this integeration length for this trial 
        Perr(np.ndarray): (nTrl,nInt : float) compute probability of error for this integeration length and trial
    '''
    if objIDs is None:
        objIDs = np.arange(Fy.shape[-1])   
    if nInt is None:
        nInt = [min(Fy.shape[-2], 30), 25]
    if not hasattr(nInt,'__iter__'):
        Int = (nInt,nInt)
    if Fy is None:
        return 1, 1, None, None, None, None, -1, 1
    if Fy.ndim > 3:
        Fyshape = Fy.shape
        Fy = np.reshape(Fy, (np.prod(Fy.shape[:-2]),)+Fy.shape[-2:]) # ((nM*nTrl),nEp,nY)

    # remove trials with no-true-label info
    keep = np.any(Fy[:, :, objIDs == 0], (-2, -1)) # [ nTrl ]
    if not np.all(keep):
        Fy = Fy[keep, :, :]
        print('Discarded %d trials without true-label info'%(sum(np.logical_not(keep))))
        if not any(keep):
            print('No trials with true label info!')
    
    # get the points at which we compute performances
    if len(nInt) < 3:
        if nInt[0] > 0 and nInt[0] < Fy.shape[-2]: #  number steps
            integerationLengths = np.linspace(0, Fy.shape[-2], min(Fy.shape[-2],nInt[0] + 1), dtype=int, endpoint=True)
        elif nInt[0] < 0: # step sized
            integerationLengths = np.arange(0, Fy.shape[-2]+1, min(Fy.shape[-2],-nInt[0]), dtype=int)
        else: 
            integerationLengths = np.arange(0, Fy.shape[-2]+1, dtype=int)
        integerationLengths = integerationLengths[1:]
    else:
        integerationLengths = nInt
    #print("intlen={}".format(integerationLengths))
        
    Yerr, Perr, aveProbErr, aveProbErrEst = compute_decoding_curve(Fy,objIDs,integerationLengths, **kwargs)

    stopPerrThresh,stopYerr = compute_stopping_curve(nInt,integerationLengths,Perr,Yerr)
    
    print(print_decoding_curve(integerationLengths,aveProbErr,aveProbErrEst,stopYerr,stopPerrThresh))
    
    return integerationLengths,aveProbErr,aveProbErrEst,stopYerr,stopPerrThresh, Yerr, Perr


def compute_decoding_curve(Fy:np.ndarray, objIDs, integerationLengths, **kwargs):
    """compute the decoding curves from the given epoch+output scores in Fy

    Args:
        Fy ([np.ndarray]): (nTrl,nEp,nY) per-epoch output scores
        objIDs ([float]): (nY,) the objectIDs for the outputs in Fy
        integerationLengths ([type]): (nInt,) a list of integeration lengths to compute peformance at

    Returns:
        Yerr(np.ndarray): (nTrl,nInt : bool) flag if estimate was correct at this integeration length for this trial 
        Perr(np.ndarray): (nTrl,nInt : float) compute probability of error for this integeration length and trial
        aveProbErr: (nInt: float) average error probablility at this integeration length
        aveProbErrEst(np.ndarray): (nInt: float) averaget estimated error probability for this integeration length
    """    
    Yidx=np.zeros((Fy.shape[0], len(integerationLengths))) # (nTrl,nInt)
    Yest=np.zeros((Fy.shape[0], len(integerationLengths))) # (nTrl,nInt)
    Perr=np.zeros((Fy.shape[0], len(integerationLengths))) # (nTrl,nInt)

    print("Int Lens:", end='')
    for li,nep in enumerate(integerationLengths):
        Yidxli,Perrli,_,_,_=decodingSupervised(Fy[:, :nep, :], **kwargs)
        # BODGE: only use result from last decision point
        Yidx[:,li]=Yidxli[:,-1]
        Perr[:,li]=Perrli[:,-1]
        # convert from Yidx to Yest, note may be invalid = -1
        Yest[:,li]=[ objIDs[yi] if yi in objIDs else -1 for yi in Yidxli[:,-1] ]
        print('.',end='',flush=True)
    print("\n")

    Yerr = Yest!=0 # (nTrl, nEp)
    aveProbErr   =np.mean(Yerr,0) #(nEp)
    aveProbErrEst=np.mean(Perr,0) #(nEp)
    return (Yerr,Perr,aveProbErr,aveProbErrEst)


def compute_stopping_curve(nInt,integerationLengths,Perr,Yerr):
    nthresh=nInt[1] if len(nInt)>1 else nInt[0]
    if nthresh < 0:
        nthresh=20
    #thresholds =linspace(min(Perr(:)),max(Perr(:)),nthresh)'; #set thresho
    thresholds=1 - np.exp(np.linspace(-0.25, -5, nthresh))
    thresholds=np.append(np.linspace(thresholds[0]*.2,thresholds[0]*.8,3),thresholds,0) # [ nThresh ]
    perrstopiYerr=np.zeros((Perr.shape[0],len(thresholds),4)) # (nTrl,nThresh,4) [ 4 x nThresh x nTrl ]
    for trli in range(Perr.shape[0]):
        tmp=Perr[trli,:]
        for ti,thresh in enumerate(thresholds):
            stopi=np.argmax(tmp < thresh) # returns 1st time test is true
            if not tmp[stopi]<thresh: 
                stopi = Perr.shape[1]-1 # set to end trial if didn't pass
            perrstopiYerr[trli,ti,:]=[thresh,Perr[trli,stopi],integerationLengths[stopi],Yerr[trli,stopi]]
    aveThreshPerrIntYerr=np.mean(perrstopiYerr,0) # (nThresh,4) [4 x nThresh] average stopping time for each threshold
    # BODGE: map into integeration lengths to allow use the same ploting routines

    # threshold with closes average stopping time
    mi=np.argmin(np.abs(integerationLengths[:,np.newaxis]-aveThreshPerrIntYerr[:,2:3].T),1) # [ nInt ] 

    stopThresh=aveThreshPerrIntYerr[mi,0]
    stopPerrThresh=aveThreshPerrIntYerr[mi,1]
    stopYerr=aveThreshPerrIntYerr[mi,3]
    return stopPerrThresh,stopYerr

    
def print_decoding_curve(integerationLengths,aveProbErr,aveProbErrEst=None,stopYerr=None,stopPerrThresh=None):
    MINSCALEPERR=0.1
    
    s=''    
    # get set lengths to print performance for
    logIdx=np.linspace(0,len(integerationLengths)-1,min(len(integerationLengths),9),dtype=int,endpoint=True)
    if  len(logIdx)>1 :
        logIdx=logIdx[1:]

    #print("logidx={}".format(logIdx))
    # make a formated summary string
    # area under decoding curve (weighted)
    audc=100.0*np.mean(aveProbErr)# ** 0.6)
    s+='%18s  '%('IntLen') + " ".join(['%4d '%(i) for i in integerationLengths[logIdx]]) + "\n"
    s+='%18s  '%('Perr')   + " ".join(['%4.2f '%(i) for i in aveProbErr[logIdx]])
    s+='  AUDC %4.1f'%(audc) + "\n"

    # PERREST
    if not aveProbErrEst is None:
        # sum squared perr estimate error
        psae=100.0*np.mean(np.abs(aveProbErrEst - aveProbErr) / np.maximum(aveProbErr,MINSCALEPERR))
        
        s+='%18s  '%('Perr(est)')+ " ".join(['%4.2f '%(i) for i in aveProbErrEst[logIdx]])    
        s+='  PSAE %4.1f'%(psae) + '\n'    

    # STOPPING CURVE
    if not stopYerr is None:
        # area under stopping curve
        ausc=100.0*np.mean(stopYerr) 
        # sum-squared error in stopping estimate
        ssae=100.0*np.mean(np.abs(stopYerr - stopPerrThresh) / np.maximum(stopYerr,MINSCALEPERR))
        s+="%18s  "%("StopErr") + " ".join(['%4.2f '%(i) for i in stopYerr[logIdx]])
        s+="  AUSC %4.1f"%(ausc) + "\n"
        s+='%18s  '%('StopThresh(P)') + " ".join(['%4.2f '%(i) for i in stopPerrThresh[logIdx]])
        s+="  SSAE %4.1f"%(ssae) + "\n"
        
    return s



def plot_decoding_curve(integerationLengths, aveProbErr, *args):
    ''' plot the decoding curve '''
    import matplotlib.pyplot as plt

    if aveProbErr.ndim > 1:
        # multiple datasets
        plt.plot(integerationLengths.T,aveProbErr.T)
        plt.plot(np.mean(integerationLengths,0), np.mean(aveProbErr,0), 'k', linewidth=5, label="mean")
        plt.title('Decoding Curve (nDatasets={})'.format(aveProbErr.shape[0]))
    
    else:
        # single dataset
        plt.plot(integerationLengths.T,aveProbErr.T,'.-',label='avePerr')
        if len(args)>=7-2:
            # plot the trialwise estimates, when is single subject
            Yerr = args[5-2] #(nTrl,nInt), flag if was right or not
            Perr = args[6-2].copy() #(nTrl,nInt)
            Perr[Yerr==True]=np.NaN # disable points where it was in error
            # est when was correct
            plt.plot(integerationLengths.T,Perr[0,:].T,'.',markerfacecolor=(0,1,0,.2),markeredgecolor=(0,1,0,.2),label='Perr(objID == 0)')
            plt.plot(integerationLengths.T,Perr.T,'.',markerfacecolor=(0,1,0,.2),markeredgecolor=(0,1,0,.2))
            # est when incorrect..
            Perr = args[6-2].copy() #(nTrl,nInt)
            Perr[Yerr==False]=np.NaN # disable points where it was in error
            plt.plot(integerationLengths.T,Perr[0,:].T,'.', markerfacecolor=(1,.0,.0,.2), markeredgecolor=(1,.0,.0,.2),label='Perr(objID ~= 0)')
            plt.plot(integerationLengths.T,Perr.T,'.', markerfacecolor=(1,.0,.0,.2), markeredgecolor=(1,.0,.0,.2))
            plt.title('Decoding Curve (nTrl={})'.format(Yerr.shape[0]))

    plt.ylim((0,1))
    plt.xlabel('Integeration Length (samples)')
    plt.ylabel('Perr')
    plt.legend()
    plt.grid(True)

def testcase():
    import numpy as np
    Fy=np.random.standard_normal((10,100,50))
    Fy[:,:,0]=Fy[:,:,0] + 0.3
    from decodingCurveSupervised import decodingCurveSupervised
    (dc)=decodingCurveSupervised(Fy)
    plot_decoding_curve(*dc)

    sFy = np.cumsum(Fy,-2)
    Yi = np.argmax(sFy,-1)
    audc=np.sum((Yi==0).ravel())/Yi.size
    print("1-Audc_score={}".format(1-audc))
    
    # test with multiple lines plotting
    il=np.tile(dc[0][np.newaxis,:],(4,1))
    pe=np.tile(dc[1][np.newaxis,:],(4,1))
    pe= pe+np.random.standard_normal(pe.shape)*.1 # add some noise
    plot_decoding_curve(il,pe)
    

if __name__=="__main__":
    testcase()



#@function
import numpy as np
def normalizeOutputScores(Fy, validTgt=None, badFyThresh=4,
                          centFy=True, detrendFy=False, 
                          nEpochCorrection=0,
                          minDecisLen=0, maxDecisLen=0,
                          bwdAccumulate=False,
                          priorsigma=None):
    '''
    normalize the raw output scores to feed into the Perr computation
    Inputs:
      Fy   - (nM,nTr,nEp,nY) [#Y x #Epoch x #Trials x nModels]:float
      validTgt - (nM,nTr,nY) [ #Y x #Trials x nModels ]:bool indication of which outputs are used in each trial
      badFyThresh - threshold for detction of bad Fy entry, in std-dev
      centFy   - bool, do we center Fy before computing  *important*  (true)
      nEpochCorrection - int, number of epochs to use a base for correction of number epochs in the sum.
               basically, nEpoch < nEpochCorrection have highly increased Perr, so unlikely to be selected
      minDecisLen - int, number of epochs to use as base for distribution of time-based decision points.
                         i.e. decisions at, [1,2,4,...2^n]*exptDistDecis
                    OR: minDecisLen<0 => decision point every abs(minDeicsLen) epochs
      maxDecisLen   - maximum number of epochs for a decision
      bwdAccumulate - [bool], accumulate data backwards from last epoch gathered
      prior_sigma (float,float) - prior estimate of the variance and it's equivalent samples weight (sigma,N)
    Outputs:
      ssFy - (nM,nTr,nDecis,nY) [#Y x #decisPts x #Trials x #Models] scaled summed scores
      sFy_scale - (nM,nTr,nDecis,1) [1 x #decisPts x #Trails x #Models] the normalization scaling factor
      Nep - (nDecis) [ #decisPts x 1 ] number of epochs included in each score
      nEp,nY - (nTrl) [ 1 x #Trials ] the detected number epoch/output for each trial
    Copyright (c) MindAffect B.V. 2018
    
    '''
    normSum = True
    if Fy is None or Fy.size == 0:
        ssFy = np.zeros(Fy.shape[:-2]+(1,Fy.shape[-1]))
        return ssFy, None, 0, None, None

    # compress out the model dimension
    Fyshape = Fy.shape # (nM,nTrl,nEp,nY)
    if Fy.ndim > 3:
        Fy = np.reshape(Fy, (np.prod(Fy.shape[:-2]), )+Fy.shape[-2:]) # ((nM*nTrl), nEp, nY) [ nY x nEp x (nTrl*nM*..)]
        if not validTgt is None : # (nM, nTr, nY)
            validTgt = np.reshape(validTgt, (np.prod(validTgt.shape[:-1]), ) + validTgt.shape[-1:]) # ((nM*nTrl), nY) [ nY x (nTrl*nM)]
    if Fy.ndim < 3: # ensure has trial dim
        Fy = Fy[np.newaxis, :, :]

    # get info on the number of valid Epochs or Outputs in each trial
    validFy = Fy != 0
    nY, nEp, validTgt = get_valid_epochs_outputs(Fy, validTgt)
    maxnEp = np.max(nEp.ravel())

    if maxnEp < 1: # guard no data to analyse
        ssFy = np.zeros(Fy.shape[:-2]+(1,Fy.shape[-1]))
        return ssFy, None, 0, None, None

    # estimate the points at which we may make a decision
    decisIdx = np.array([maxnEp-1], dtype=int) # default to single decision at max-length
    if minDecisLen > 0 and minDecisLen < maxnEp:
        # exp-distributed decision points
        nstep = int(np.ceil(np.log((maxnEp-1) / minDecisLen) / np.log(2)))
        decisIdx = [(2**i)*minDecisLen for i in range(nstep+1)]
        decisIdx = np.minimum(np.maximum(decisIdx, minDecisLen), maxnEp-1)
    elif minDecisLen < 0:
        # linearly distributed decision points
        decisIdx = np.arange(-minDecisLen-1, maxnEp-1, -minDecisLen)
        decisIdx = np.append(decisIdx, maxnEp-1)
    else :
        # minDecisLen = 0 -> max-length
        pass
    #print("decisIdx={}".format(decisIdx))

    # Limit the max window size to test
    if maxDecisLen > 0 and maxDecisLen >= decisIdx[0]:
        decisIdx = decisIdx[decisIdx <= maxDecisLen]

    # For computational efficiency remove data outside the max decision points
    if np.max(decisIdx[-1]) < min(Fy.shape[-2]*0.7, Fy.shape[-2]-100): # (nM, nTrl, nEp, nY)
        maxLen = int(np.max(decisIdx[-1]))
        if bwdAccumulate:
            Fy = Fy[:, Fy.shape[-2]-maxLen:, :]
        else:
            Fy = Fy[:, :maxLen+1, :]

    # compute the summed scores
    if abs(minDecisLen) > Fy.shape[-2]:
        decisIdx = Fy.shape[-2]-1
        N = nEp[:, np.newaxis] # (nTrl, nDecis) [nDecis x nTrl] number elements in the sum
        sFy = np.sum(Fy, -2)
    else:
        if (bwdAccumulate): # (nM, nTrl, nEp, nY) # [nY, nEp, nTrl, nM]
            if not np.all(nEp == nEp[0]):
                for ti in range(Fy.shape[-3]):
                    Fy[ti, :nEp[ti], :] = Fy[ti, nEp[ti]-1::-1, :]
            else:
                Fy = Fy[:, ::-1, :]
            validFy = Fy != 0
        sFy = np.cumsum(Fy, -2)
        sFy = sFy[:, decisIdx, :]

        # (nTrl, nDecis) number points in each sum
        N   = np.cumsum(np.any(validFy,-1),-1) # (nTrl,nEp)
        N   = N[:, decisIdx]

        #N = np.tile(decisIdx[np.newaxis, :], (Fy.shape[-3], 1)) # (nTrl, nDecis) number points in each sum [ nDecis x nTrl ]
        #for ti, nti in enumerate(nEp): # and limit to trial length
        #    N[ti, :] = np.minimum(nti, N[ti, :])

    # compute the noise scale to transform to standardized units
    sigma2, Nsigma = estimate_Fy_noise_variance_2(Fy, decisIdx=decisIdx, centFy=centFy, detrendFy=detrendFy, priorsigma=priorsigma)

    # scale = std-deviation over outputs of smoothed summed score at each decison point
    if (normSum):
        # \sum_i N(0, sigma)) ~ N(0, sqrt(i)*sigma) 
        # TODO: use N= number non-zero entries rather than just length..
        sFy_scale = np.sqrt(sigma2*np.maximum(.01,N)) # [ nDecis x nTrl ]
    else:
        sFy_scale = np.sqrt(sigma2)
    #plt.clf();plt.subplot(221);plt.plot(Fy[:, :, 0].T);plt.title('Fy');plt.subplot(222);plt.plot(fFy[:, :, 0].T);plt.title('fFy');plt.subplot(223);plt.plot(sFy[:, :, 0].T);plt.title('sFy');plt.subplot(224);plt.plot(sfFy[:, :, 0].T);plt.title('sfFy');plt.show()
    #sFy_scale_emp=np.std(sfFy[:, :, 0], 0)
    #print(np.mean((sFy_scale_emp[:, np.newaxis]/sFy_scale).ravel()))
    #plt.subplot(222);plt.plot(np.sqrt(sigma2), 'k');plt.plot(np.std(fFy[:, :, 0], 0), 'k.'); plt.subplot(224);plt.plot(sFy_scale, 'k');plt.plot(np.std(sfFy[:, :, 0], 0), 'k.');plt.show()
    #plt.clf();plt.subplot(211);plt.plot(sFy[:, :, 0].T);plt.plot(sFy_scale[:, 0], 'k');plt.show()
    #print('sigma2 = {}'.format(np.mean(sigma2.ravel())))


    # correction factor for sampling bias in the std
    # estimation of of the correction factors to
    # from: https://en.wikipedia.org/wiki/Unbiased_estimation_of_standard_deviation
    # E[sigma]=c4(n)*sigma -> sigma = E[sigma]/c4(n)
    # where cf is the correction for the sampling bias in the estimator
    if nEpochCorrection is not None and nEpochCorrection > 0 :
        cf = c4(N/np.maximum(1, nEpochCorrection)) 
        #cf = c4(N/np.maximum(1, nEpochCorrection))**2 # too agressive 
        # include the multiple comparsiosn correction factors
        sFy_scale = sFy_scale / cf

    # apply the normalization to convert to z-score (i.e. unit-noise)
    sFy_scale[sFy_scale == 0] = 1
    ssFy = sFy/sFy_scale[:, :, np.newaxis]

    # reverse the time reversal for backward accumulation    
    if bwdAccumulate:
        ssFy = ssFy[:, ::-1, :]
        sFy_scale = sFy_scale[:, ::-1, :]
        N = N[::-1]

    if len(Fyshape)>3 : # convert back to have model dimension
        ssFy      = np.reshape(ssFy, (Fyshape[:-2]+ssFy.shape[-2:]))
        sFy_scale = np.reshape(sFy_scale, (Fyshape[:-2]+sFy_scale.shape[-1:]))
        nEp       = np.reshape(nEp, Fyshape[:-2])
        nY        = np.reshape(nY, Fyshape[:-2])

    return ssFy, sFy_scale, decisIdx, nEp, nY

def get_valid_epochs_outputs(Fy,validTgt=None):
    """get the number valid epoch and outputs from a zero-padded Fy set

    Args:
        Fy ([np.ndarray]): (nTrl,nEp,nY) the output scores
        validTgt ([np.ndarray]): (nTrl,nY) flag if this output is used in this trial 

    Returns:
        nY ([np.ndarray]): (nTrl,) number valid outputs per trial
        nEp ([np.ndarray]): (nTrl,) number valid epochs per trial
        validTgt ([np.ndarray]): (nTrl,nY) flag if this output is used in this trial 
    """    
    if validTgt is None : # (nM, nTr, nEp, nY)
        validTgt = np.any(Fy != 0, -2) # (nM, nTr, nY) [ nY x nTrl ]
    nY  = np.sum(validTgt, -1) # number active outputs in this trial (nM*nTrl) [ nTrl x nM ]
    nEp = np.ones((Fy.shape[-3]), dtype=int) # number active epochs in each trial (nM*nTrl) [ nTrl x nM ]
    for ti in range(Fy.shape[-3]):
        tmp = np.flatnonzero(np.any(Fy[ti, :, :] != 0, axis=-1))
        nEp[ti] = tmp[-1] if tmp.size > 0 else 0 # num active epochs/samples
    return nY, nEp, validTgt

def filter_Fy(Fy, filtLen=None, B=None):
    # apply the temporal smoothing filter to the raw scores
    # i.e. low-pass,  to remove short-term temporal 
    # (anti)-correlations which add a lot of noise to the scale
    # estimation # (nM, nTrl, nEp, nY) # [nY, nEp, nTrl, nM]
    if B is None and (filtLen is None or filtLen == 1 or filtLen == 0):
        return Fy
    
    if B is None:
        filtLen = max(1, min(filtLen, Fy.shape[-2]))
        B = np.ones((filtLen))  # simple moving average filter
    else:
        filtLen = len(B)

    #B = np.array((.5,-1.5,2,-1.5,.5))/2 # simple high-pass filter
    # TODO []: how to norm B to achive properties we want?
    B = B /  np.sum(np.abs(B))
    # TODO: good multi-dimensional filter... & check correctness of the convolve
    fFy=np.zeros(Fy.shape) # (nTrl, nEp, nY) [ nY x nEp x nTrl ]
    for ti in range(Fy.shape[-3]): # trials loop
        for yi in range(Fy.shape[-1]): # outputs loop
            fFy[ti, :, yi] = np.convolve(Fy[ti, :, yi], B)[:Fy.shape[-2]] # do the smoothing
    # compute the point weighting
    wB  = np.convolve(np.ones(Fy.shape[-2]), np.abs(B))[:Fy.shape[-2]] # point weighting for each element of fFy (nEp) [ nEp,  ]
    # normalize away the point weighting,  and startup effects
    fFy = fFy/wB[np.newaxis, :, np.newaxis]
    #w=np.ones(Fy.shape[1])
    # compare raw and filtered sum
    #sfFy  = np.cumsum(fFy, -2)[:, decisIdx, :] # (nTrl, nDecis, nY) [ nY x nDecis x nTrl] sfFy(y, T) = \sum_t=0^T fFy(y, t)
    return fFy

def estimate_Fy_noise_variance_2(Fy, decisIdx=None, centFy=True, detrendFy=False, priorsigma=None,  verb=0):
    """Estimate the noise variance for Fy

    Args:
        Fy ([np.ndarray]): (nTr,nEp,nY) the output scores
        decisIdx ([type], optional): [description]. Defaults to None.
        centFy (bool, optional): flag if we should center over outputs before computing variance. Defaults to True.
        detrendFy (bool, optional): flag if we should detrend Fy before computing it's variance. Defaults to True.
        priorsigma ([type], optional): Prior estimate for the variance. Defaults to None.
    returns:
        sigma2 ([np.ndarray]): (nTr, nDecis) estimated variance per sample at each decision point
    """
    validFy = Fy != 0
    nY  = np.sum(np.any(validFy, -2), -1) # number active outputs in this trial (nM*nTrl) 
    N   = np.cumsum(np.any(validFy,-1),-1) # (nTr, nEp) number valid epochs before time-point in trial

    if decisIdx is None:
        decisIdx = np.array([Fy.shape[-2]-1],dtype=int)

    # TODO [] : generalize by all possible accumul lengths at ass poss start points
    # TODO [] : make more computationally efficient? e.g. no explicit centering
    cFy = Fy.copy()

    # compute the cumulative sum
    scFy = np.cumsum(cFy, -2) # cum-sum from start # (nTr, nEp, nY)

    # remove per-example offset (if enough outputs to do reliably)
    if centFy and np.all(nY>3):
        # center at each time point, with guard for no active outputs (when mu should == 0)
        muFy_y = np.sum(scFy, -1, keepdims=True) / nY[:, np.newaxis, np.newaxis] # mean at each time-point
        scFy = scFy - muFy_y

    # variance of the summed scores over outputs for each time point
    # if independent then this should be a constant slope increase over time.
    var2csFy = np.sum(scFy**2, -1) / np.maximum(.1,nY[:,np.newaxis]-1) # (nTr,nEp) var over outputs for each cumsum

    # normalize out the constant slope over time, to get the equivalent slope if 
    # we had N-indenpendent samples  
    nvar2csFy = var2csFy / np.maximum(.1, N) #np.arange(1,var2csFy.shape[-1]+1) # ave var per-time-step

    # compute the average of the estimated slopes for each integeration length
    muvar2csFy = np.cumsum(nvar2csFy, -1) / np.maximum(.1, N) # np.arange(1,nvar2csFy.shape[-1]+1) # ave per-stime-stamp vars before each time-point
    
    # return the ave-cumsum-var-slope at each decision length
    sigma2 = muvar2csFy[:,decisIdx] # (nTr,nDecis)
    N = N[:,decisIdx]

    # include the prior if needed
    if priorsigma is not None and priorsigma[0]>0 :
        # include the effect of the prior, sigma is weighted combo pior and data
        #  sigma'^2 = 1 / ( N_0/sigma_0^2 + N / sigma^2) 
        #           = sigma_0^2 * sigma^2 / ( N_0 sigma^2 + N * sigma_0 )
        osigma2= np.mean(sigma2)
        sigma2 = (sigma2*decisIdx + priorsigma[0]*priorsigma[1]) / ( decisIdx + priorsigma[1] )
        #print('sigma2 = {}  prior={} -> {}'.format(osigma2,priorsigma,np.mean(sigma2)))
    
    return sigma2, N


def estimate_Fy_noise_variance(Fy, decisIdx=None, centFy=True, detrendFy=False, priorsigma=None,  verb=0):
    """Estimate the noise variance for Fy

    Args:
        Fy ([np.ndarray]): (nTr,nEp,nY) the output scores
        decisIdx ([type], optional): [description]. Defaults to None.
        centFy (bool, optional): flag if we should center over outputs before computing variance. Defaults to True.
        detrendFy (bool, optional): flag if we should detrend Fy before computing it's variance. Defaults to True.
        priorsigma ([type], optional): Prior estimate for the variance. Defaults to None.
    returns:
        sigma2 ([np.ndarray]): (nTr, nDecis) estimated variance per sample at each decision point
    """
    if decisIdx is None:
        decisIdx = np.array([Fy.shape[-2]-1],dtype=int)

    sigma2 = np.zeros((Fy.shape[0],decisIdx.size),dtype=Fy.dtype)
    N      = np.zeros((Fy.shape[0],decisIdx.size),dtype=int)
    for ti in range(Fy.shape[0]):
        Fyti = Fy[ti,...].copy()
        # subset to valid entries, samples/outputs
        vi = np.any(Fyti!=0,-1) # active epochs/samples
        Fyti = Fyti[vi,:] # remove the inactive 
        vY = np.any(Fyti!=0,-2) # active outputs
        Fyti = Fyti[:,vY]

        # get the decision points in the compressed representation
        Nti = np.cumsum(vi,-1) # #active epoch at each index == index-1
        decisIdxti = Nti[decisIdx]-1 # index of the decision points in compressed Fy
        N[ti, :] = Nti[decisIdx]

        if Fyti.size == 0:
            sigma2[ti,:]=np.nan # undefined if no active points!
            continue

        if detrendFy and Fyti.shape[-2]>10:
            # center over time-points
            muFy_t = np.mean(Fyti,-2, keepdims=True)
            Fyti = Fyti - muFy_t

        # compute the cumulative sum
        sFy = np.cumsum(Fyti, -2) # cum-sum from start # (nTr, nEp, nY)

        # remove per-example offset (if enough outputs to do reliably)
        if centFy and Fyti.shape[-1]>3:
            # center at each time point, with guard for no active outputs (when mu should == 0)
            muFy_y = np.mean(sFy, -1, keepdims=True) # mean at each time-point
            sFy = sFy - muFy_y

        # variance of the summed scores over outputs for each time point
        # if independent then this should be a constant slope increase over time.
        var2sFy = np.sum(sFy**2, -1) / np.maximum(.1,sFy.shape[-1]-1) # (nTr,nEp) var over outputs for each cumsum

        # normalize out the constant slope over time, to get the equivalent slope if 
        # we had N-indenpendent samples  
        nvar2sFy = var2sFy / np.arange(1,var2sFy.shape[-1]+1) # ave var per-time-step

        # compute the average of the estimated slopes for each integeration length
        muvar2sFy = np.cumsum(nvar2sFy, -1) / np.arange(1,var2sFy.shape[-1]+1) # np.arange(1,nvar2csFy.shape[-1]+1) # ave per-stime-stamp vars before each time-point
    
        # return the ave-cumsum-var-slope at each decision length
        sigma2[ti,:] = muvar2sFy[decisIdxti] # (nTr,nDecis)
        sigma2[ti,decisIdxti<0]=np.nan # undefined if no active points!
        

    # include the prior if needed
    if priorsigma is not None and priorsigma[0]>0 :
        # include the effect of the prior, sigma is weighted combo pior and data
        #  sigma'^2 = 1 / ( N_0/sigma_0^2 + N / sigma^2) 
        #           = sigma_0^2 * sigma^2 / ( N_0 sigma^2 + N * sigma_0 )
        osigma2= np.mean(sigma2)
        sigma2 = (sigma2*N + priorsigma[0]*priorsigma[1]) / ( N + priorsigma[1] )
        #print('sigma2 = {}  prior={} -> {}'.format(osigma2,priorsigma,np.mean(sigma2)))
    
    return sigma2, N


#@function
def c4(n):
    # correction factor for bias in the standard deviation  
# from: https://en.wikipedia.org/wiki/Unbiased_estimation_of_standard_deviation
    n = np.maximum(1.0, n)
    cf = 1 - 1 / 4*(n ** - 1) - 7 / 32*(n ** - 2) - 19 / 128*(n ** - 3)
    return cf

def plot_normalizedScores(Fy, ssFy, scale_sFy, decisIdx):
    '''plot the normalized score and raw summed score for comparsion '''
    sFy = np.cumsum(Fy, -2) # (nEp, nY)
    stdsFy = np.std(sFy,-1,keepdims=True)
    #stdsFy = np.sum((sFy-np.mean(sFy,-1,keepdims=True))**2, -1) /sFy.shape[-1] / np.arange(1,sFy.shape[-2]+1)
    #stdsFy = np.cumsum(stdsFy) / np.arange(1,stdsFy.shape[-1]+1)
    print('sFy={}'.format(sFy.shape))
    import matplotlib.pyplot as plt
    plt.clf()
    plt.subplot(211)
    plt.cla()
    plt.plot(decisIdx, sFy[decisIdx, :]) 
    plt.plot(decisIdx, scale_sFy, 'k', label='scale_sFy')
    plt.plot(stdsFy, 'k.', label='std cumsum(Fy)')
    plt.legend()
    plt.title('Non-normalized sum - scale')
    plt.grid('on')
    plt.subplot(212)
    plt.cla()
    plt.plot(decisIdx, ssFy)
    stdssFy = np.std(ssFy,-1,keepdims=True)
    plt.plot(decisIdx, stdssFy[decisIdx,:],'k.',label='std ssFy')
    plt.title("scaled sum")
    plt.ylabel('ssFy'); plt.xlabel('epoch')
    plt.grid('on')
    plt.show()


#@function
def mktestFy(nY=10, nM=1, nEp=360, nTrl=100, sigstr=.5, startupNoisefrac=.5, offsetstr=10, trlenfrac=.25):
    import numpy as np
    noise = np.random.standard_normal((nM, nTrl, nEp, nY))
    noise = noise - np.mean(noise.ravel())
    noise = noise / np.std(noise.ravel())
    # add an offset to the scores...
    offset = offsetstr*np.random.standard_normal((noise.shape[-2])) # (nEp)
    noise = noise+offset[:, np.newaxis] # (nM, nTrl, nEp, nY) [ nYxnEpxnTrlxnM ]


    sigamp = sigstr*np.ones((noise.shape[-2])) # (nEp)
    # no signal at the start of the trial
    startupNoise_samp = int(nEp*startupNoisefrac)
    sigamp[:startupNoise_samp] = 0
    noise[:,:,:startupNoise_samp,:]=0

    # measure is sig + noise
    Fy = noise
    Fy[0, :, :, 0] = Fy[0, :, :, 0] + sigamp[np.newaxis, :] # (nM, nTrl, nEp, nY) [nEp x nTrl]
    
    # make variable length trials,  and per-trial offset
    nEp = np.zeros((Fy.shape[-3]), dtype=int)
    for ti in range(Fy.shape[-3]):
        # trial offset
        mu = np.random.standard_normal()
        Fy[:,  ti, :, :] = Fy[:, ti, :, :] + mu
        # trial end time
        nEp[ti] = Fy.shape[-2]*(np.random.uniform()*trlenfrac + (1-trlenfrac))
        Fy[:, ti, nEp[ti]:, :] = 0
    return Fy, nEp

def testcase():
    detrendFy=True
    centFy=True

    from normalizeOutputScores import mktestFy,  normalizeOutputScores
    import matplotlib.pyplot as plt
    Fy, nEp = mktestFy(sigstr=1) #(nM, nTrl, nEp, nY)
    Fy = Fy[0,...] # (nTrl,nEp,nY)
    # Introduce temporal and spatial sparsity like real data
    Fy = Fy * (np.random.standard_normal((1,Fy.shape[-2],Fy.shape[-1]))>0).astype(np.float)
    Fy[:,:20,:] = 0 # block zero at start

    Fy[:, 0::3, :] = 0 # only 50% active
    Fy[:, 1::3, :] = 0 # only 50% active
    oFy=Fy.copy()

    print("Fy={}".format(Fy.shape))
    
    # visualize all trials true-target normalized scores
    ssFy, scale_sFy, decisIdx, nEp, nY = normalizeOutputScores(Fy, minDecisLen=-1, nEpochCorrection=0, centFy=centFy, detrendFy=detrendFy)
    print('ssFy={} scale_sFy={}'.format(ssFy.shape,scale_sFy.shape))
    #%matplotlib
    plt.figure(1)
    plot_normalizedScores(oFy[0,...],ssFy[0,:,:],scale_sFy[0,:],decisIdx)

    # introduce temporal correlations and visualize
    Fy = filter_Fy(Fy, filtLen=10)
    ssFy, scale_sFy, decisIdx, nEp, nY = normalizeOutputScores(Fy, minDecisLen=-1, nEpochCorrection=0, centFy=centFy, detrendFy=detrendFy)
    print('pre-filtered ssFy={}'.format(ssFy.shape))
    #%matplotlib
    plt.figure(1)
    plt.suptitle("filtered {}".format(10))
    plot_normalizedScores(Fy[0,:,:],ssFy[0,:,:],scale_sFy[0,:],decisIdx)

    # introduce temporal correlations and visualize
    Fy = filter_Fy(Fy, B=np.array([1,0,0,-1]))
    ssFy, scale_sFy, decisIdx, nEp, nY = normalizeOutputScores(Fy, minDecisLen=-1, nEpochCorrection=0, centFy=centFy, detrendFy=detrendFy)
    print('pre-filtered ssFy={}'.format(ssFy.shape))
    #%matplotlib
    plt.figure(1)
    plt.suptitle('anti-correlated')
    plot_normalizedScores(Fy[0,:,:],ssFy[0,:,:],scale_sFy[0,:],decisIdx)

if __name__=="__main__":
    testcase()



import os
import numpy as np
from scipy.io import loadmat, whosmat
from mindaffectBCI.decoder.utils import block_randomize, butter_sosfilt, upsample_codebook, lab2ind
from mindaffectBCI.decoder.preprocess import extract_envelope

SSVEP_FREQS=(12, 8.57, 6.67, 5.45)
ERP_STIM_DUR = 80/1000 # 80ms
MI_STIM_DUR = 4 # 3s
SSVEP_STIM_DUR = 4

def load_openBMI(datadir, sessdir=None, sessfn=None, fs_out=60, stopband=((45,65),(0,1),(25,-1)), CAR=False, verb=1, trlen_ms=None, offset_ms=(0,0), ppMI=True):
    
    if offset_ms is None:
        offset_ms = (0, 0)

    # load the data file
    Xfn = datadir
    if sessdir:
        Xfn = os.path.join(Xfn, sessdir)
    if sessfn:
        Xfn = os.path.join(Xfn, sessfn)
    sessdir = os.path.dirname(Xfn)
    fn = os.path.basename(Xfn)
    print("Loading: {}".format(Xfn))
    data = loadmat(Xfn)

    if 'EEG_SSVEP_train' in data:
        # TODO[] : merge training and test
        data = data['EEG_SSVEP_train']
    elif 'EEG_ERP_train' in data:
        data = data['EEG_ERP_train']
    elif 'EEG_MI_train' in data:
        data = data['EEG_MI_train']

    def squeeze(v):
        while v.size == 1 and v.ndim > 0:
            v = v[0]
        return v

    fs = squeeze(data['fs'])
    ch_names = [d for d in data['chan'][0,:]]
    X = squeeze(data['x']) # (nSamp,d)
    X = np.asarray(X, order='C', dtype='float32')
    print("X={}".format(X.shape),flush=True)
    trl_idx = np.asarray(squeeze(data['t']).ravel(), dtype=int) # (nTrl)
    lab = np.asarray(squeeze(data['y_dec']).ravel(), dtype=int) # (nTrl) label for each trial

    if fs>250: # pre-resample....
        rr = int(fs//250)
        print("pre-downsample : {} -> {}".format(fs,fs/rr))
        X = X [::rr,...]
        trl_idx = trl_idx//rr
        fs = fs/rr
        
    if 'ERP' in fn:
        ep_idx, ep_trl_idx = get_trl_ep_idx(trl_idx, fs*2)
        trl_idx = ep_idx[:,0]
        # make lab the right shape also
        lab0 = lab
        lab = np.zeros(ep_idx.shape,dtype=int)
        for ei, ti_idx in enumerate(ep_trl_idx):
            lab[ei,:len(ti_idx)] = lab0[ti_idx]
        # auto-determine the  right trial length
        if trlen_ms is None:
            trlen_samp = np.max(ep_idx[:,-1]-ep_idx[:,0])
            trlen_ms = trlen_samp * 1000 / fs
    else:
        ep_idx = None
        if trlen_ms is None:
            if 'SSVEP' in Xfn:
                trlen_ms = SSVEP_STIM_DUR * 1000
            elif 'MI' in Xfn:
                trlen_ms = MI_STIM_DUR * 1000
        
    # delete the data variable to free the ram
    del data
    
    # preprocess -> spectral filter, in continuous time!
    if stopband is not None:
        if verb > 0:  print("preFilter: {}Hz".format(stopband))
        X, _, _ = butter_sosfilt(X,stopband,fs)

    if CAR:
        if verb>0 : print("CAR")
        X = X - np.mean(X,-1,keepdims=True)

    # pre-process -> band-power....
    # N.B. *before* slicing to avoid startup artifacts.
    if "MI" in fn and ppMI: # make the target array
        # map X to (log)power in MI relevant frequency bands,
        # N.B. filter-band of **STOPBANDS**
        filterbank=(((0,8),(16,-1)),((0,18),(30,-1))) # relevant frequency bands, mu~=10-14, beta~=20-25 
        env_stopband=(1,-1) # N.B. should be < min(filter-freq)/2 to smooth rectified
        print("map to envelope in bands {}".format(filterbank))
        X = extract_envelope(X,fs,stopband=None,whiten=True,filterbank=filterbank,log=False,env_stopband=env_stopband)
        # update the channel label to include band info
        ch_names = ["{} @{}hz".format(c[0],f[0]) for f in filterbank for c in ch_names]
        if verb>1: print("ch_names={}".format(ch_names))    
    
    # slice X
    trlen_samp = int(trlen_ms * fs / 1000)
    offset_samp = [int(o*fs/1000) for o in offset_ms] # relative to start/end (0,trlen_samp)
    bgnend_samp = (offset_samp[0], trlen_samp+offset_samp[1]) # start end slice window
    xlen_samp = bgnend_samp[1]-bgnend_samp[0]
    print("xslice_samp =[{} - {}] @ {}Hz".format(bgnend_samp[0], bgnend_samp[1],fs))
    Xraw = X
    X = np.zeros((len(trl_idx), xlen_samp, Xraw.shape[-1]), dtype='float32') # (nTrl,nsamp,d)
    for ti, si in enumerate(trl_idx):
        X[ti, :, :] = Xraw[si+bgnend_samp[0]:si+bgnend_samp[1], :]
    del Xraw
        
    # extract the reference signals
    if 'SSVEP' in fn:
        # convert lab+code into Y + samp-times,  where 1st row is always the true label
        if offset_samp[0] > 0 or offset_samp[1] < 0:
            # otherwise more complex to work out how to insert
            raise ValueError("Only offset_ms which padds slice currently supported")
        else:
            cb_idx = slice(-offset_samp[0], -offset_samp[1] if offset_samp[1] > 0 else None)

        # make ssvep codebook 
        cb = make_ssvep_ref_signals(trlen_samp, freqs=SSVEP_FREQS, fs=fs, phases=None) # (nY,nsamp)
        Y  = np.zeros((len(lab), X.shape[1], cb.shape[1]+1), dtype='float32') # (nTr, nSamp, nY)
        for ti, l in enumerate(lab):
            Y[ti, cb_idx, 0]  = cb[:, l-1] # copy in true label
            Y[ti, cb_idx, 1:] = cb # copy in other outputs
        
    elif 'ERP' in fn:
        # permute to make other possible entries
        cb_true = lab == 2 # target=flash (nTrl, nEp)
        cb_true = cb_true[:, :, np.newaxis, np.newaxis] # (nTrl,nEp,nY,e)
        # make 35 virtual targets
        cb = block_randomize(cb_true, 35, axis=-3)
        cb = np.concatenate((cb_true, cb), axis=-2) # [ ..., nY+1]
        # make into a sample rate label set
        Y = upsample_codebook(xlen_samp, cb, ep_idx, fs*ERP_STIM_DUR, offset_samp)
        # BODGE: strip the  feature  dim again..
        Y = Y[...,0]
        
    elif "MI" in fn: # make the target array
        # permute to make other possible entries
        cb_true, lab2class = lab2ind(lab) # (nTrl, e) # feature dim per class
        cb_true = cb_true[:,np.newaxis,:] # (nTrl,1,e)
        cb_all  = np.eye(cb_true.shape[-1]) #(nvirt, e)
        cb_all  = np.tile(cb_all,(cb_true.shape[0], 1, 1)) # (nTrl,nvirt,e)
        cb = np.append(cb_true, cb_all, axis=-2) # (nTrl, nvirt+1, e)
        cb = cb[:,np.newaxis,:,:] # (nTrl,1,nvirt+1,e)
        # make into a sample rate label set
        Y = upsample_codebook(xlen_samp, cb, None, fs*MI_STIM_DUR, offset_samp) #(nTrl,nSamp,nY,e)

        
    # preprocess -> downsample
    resamprate = int(fs/fs_out)
    if resamprate > 1:
        if verb > 0:
            print("resample: {}->{}hz rsrate={}".format(fs, fs/resamprate, resamprate))
        X = X[:, ::resamprate, :] # decimate X (trl, samp, d)
        Y = Y[:, ::resamprate, :] # decimate Y (trl, samp, y)
        fs = fs/resamprate

    stimTimes = None # non-sliced output
    
    # make coords array for the meta-info about the dimensions of X
    coords = [None]*X.ndim
    coords[0] = {'name':'trial','coords':trl_idx, 'lab':lab }
    coords[1] = {'name':'time','unit':'ms', \
                 'coords':np.linspace(offset_ms[0],trlen_ms+offset_ms[1],X.shape[1]), \
                 'fs':fs}
    coords[2] = {'name':'channel','coords':ch_names}
    # return data + metadata
    return (X, Y, coords)


def make_ssvep_ref_signals(nsample, freqs, phases=None, fs=1, binarize=True):
    '''make periodic reference signals'''
    ref = np.zeros((nsample, len(freqs)), dtype='float32')
    for i, freq in enumerate(freqs):
        isi = fs/freq # period in samples
        phase = 0 if phases is None else phases[i]
        ref[:, i] = np.sin((np.arange(nsample)/isi+phase)*2*np.pi)
        if binarize: # put event at phase==0
            ref[:, i] = ref[:, i] > 0
    return ref

def get_trl_ep_idx(trl_idx, iti):
    ''' convert set stimulus times into trials and epochs '''
    # large gap indicates end of this trial
    tr_lims = np.flatnonzero(np.diff(trl_idx) > iti)+1
    tr_lims = np.concatenate(([0], tr_lims, [len(trl_idx)]))
    # insert trial epoch indices into the epoch indices array 
    nEp = np.max(np.diff(tr_lims))
    ep_idx = np.zeros((len(tr_lims)-1, nEp), dtype=int) # (nTrl, nEp)
    ep_trl_idx = [None]*(len(tr_lims)-1) # (nTrl, nEp), list of lists
    for ei in range(0, len(tr_lims)-1):
        tmp = range(tr_lims[ei],tr_lims[ei+1])
        ep_trl_idx[ei] = list(tmp)
        ep_idx[ei, :len(tmp)] = trl_idx[tmp]
    return (ep_idx, ep_trl_idx)


def testcase():
    import sys

    if os.path.isdir('D:\external_data'):
        sessfn = 'D:\external_data\gigadb\openBMI\session1\s1\sml_sess01_subj01_EEG_SSVEP.mat' 
        sessfn = 'D:\external_data\gigadb\openBMI\session1\s1\sml_sess01_subj01_EEG_ERP.mat'
        sessfn = 'D:\external_data\gigadb\openBMI\session1\s2\sml_sess01_subj02_EEG_MI.mat'
    else:
        sessfn = '/home/jadref/data/bci/external_data/gigadb/openBMI/session1/s2/sml_sess01_subj02_EEG_MI.mat'

    # command-line, for testing
    if len(sys.argv) > 1:
        sessfn = sys.argv[1]

    from load_openBMI import load_openBMI
    X, Y, coords = load_openBMI(sessfn, CAR=True, offset_ms=(-400,1000), sessfn=sessfn, fs_out=60, stopband=((0,1),(30,-1)))
    fs = coords[1]['fs']
    # CAR=False,fs_out=60,stopband=((0,3),(29,-1)),rcond=1e-3 : audc=36 Perr[-1]=.30
    # CAR=True,fs_out=60,stopband=((0,3),(29,-1)),rcond=1e-3 : audc=36 Perr[-1]=.30
    # CAR=True,fs_out=60,stopband=((0,3),(29,-1)),rcond=1e-8 : audc=36 Perr[-1]=.30

    
    if 'SSVEP' in sessfn:
        tau_ms = 300
        evtlabs=('re')
        rank=1
    elif 'ERP' in sessfn:
        tau_ms = 700
        evtlabs = ('re','ntre')
        rank=10
    elif 'MI' in sessfn:
        tau_ms = 20
        evtlabs = None
        rank=5

    # visualize the dataset
    from analyse_datasets import debug_test_dataset
    debug_test_dataset(X,Y,coords, rank=rank, evtlabs=evtlabs, tau_ms=tau_ms, offset_ms=0, model='cca', rcond=1e-2)
    
if __name__=="__main__":
    testcase()




from scipy.io import loadmat
from mindaffectBCI.decoder.utils import window_axis, block_randomize, butter_sosfilt
import numpy as np
def load_mTRF_audio(datadir, regressor='envelope', ntrl=15, stopband=((45,65),(0,.5),(15,-1)), fs_out=60, nvirt_out=30, verb=1):
    d = loadmat(datadir)
    X = d['EEG'] # (nSamp,d)
    Y = d[regressor] # (nSamp,e)
    Y = Y[:, np.newaxis, :] # (nSamp, nY, e)
    fs = d['Fs'][0][0]
    if fs_out is  None:
        fs_out = fs

    # preprocess -> spectral filter, in continuous time!
    if stopband is not None:
        if verb > 0:
            print("preFilter: {}Hz".format(stopband))
        X,_,_ = butter_sosfilt(X, stopband, fs, axis=-2)
    
    # generate artificial other stimulus streams, for testing
    Y_test = block_randomize(Y, nvirt_out, axis=-3, block_size=Y.shape[0]//ntrl//2)
    Y = np.concatenate((Y, Y_test), -2) # (nSamp, nY, e)
    
    # slice X,Y into 'trials'
    if  ntrl > 1:
        winsz = X.shape[0]//ntrl
        X = window_axis(X, axis=0, winsz=winsz, step=winsz) # (ntrl,nSamp,d)
        Y = window_axis(Y, axis=0, winsz=winsz, step=winsz) # (nTrl,nSamp,nY,e)
    else:
        X = [np.newaxis, ...]
        Y = [np.newaxis, ...]

    # preprocess -> downsample
    resamprate = int(fs/fs_out)
    if resamprate > 1:
        if verb > 0:
            print("resample: {}->{}hz rsrate={}".format(fs, fs/resamprate, resamprate))
        X = X[:, ::resamprate, :] # decimate X (trl, samp, d)
        Y = Y[:, ::resamprate, :] # decimate Y (trl, samp, y)
        fs = fs/resamprate
        
    # make meta-info
    coords = [None]*X.ndim
    coords[0] = {'name':'trial'}
    coords[1] = {'name':'time', 'fs':fs, 'units':'ms', 'coords':np.arange(X.shape[1])*1000/fs}
    coords[2] = {'name':'channel', 'coords':None}

    return (X, Y, coords)



from .ganglion import OpenBCIGanglion
import asyncore # needed for the openBCI server
import numpy as np
from time import time,sleep
import socket

# add utopiaclient to path and import
import sys, os
from mindaffectBCI import utopiaclient

SCALE_FACTOR_EEG = 1000000.0/(1.2*8388607.0*1.5*51.0) #uV/count
SCALE_FACTOR_AUX = 0.032
LOGINTERVAL_S=2

lastdata=None # last valid data seen
lastsamp=None # last sample id processed
lastcount=None # number of times we have seen this sample id -- for multi-sample packets
t0=None # start time
nextLogTime=None # time print next log message
nSamp=0 # number samples processed so far
ftBlockSample=200/20 # minimum number samples in each block sent, so stream @20hz
nBlock=0 # number blocks sent to ft-buffer
ftBlock=[] # storage for the current block samples to go to ft-buffer
logFile=None # "ganglion.log"
fSample=None
samp0  =None

def getTime_ms():
    if client:
        return client.getTimeStamp()
    return None
def getTime():
    return getTime_ms()/1000

def incsampCount(lastsample,count):
    ''' wrap arround sample incrementer, with repeate tracker '''
    if lastsample==0 :
        sample=101
        count=0
    elif count==0 :
        sample=lastsample
        count=count+1
    elif count==1 :
        if lastsample==200 or lastsample==100: # wrap around
            sample=0
            count=1 # BODGE: 0 doesn't repeat
        else : # increment sample counter
            sample=lastsample+1
            count=0
    return (sample,count)

nerr=0 # number of samples with sample underestimate
estndropped=0 # estimated magnitude of the missing samples
epsilon_err=1 # size of missing sample deadzone
thresh_nerr_s=6 # number seconds consistently to few samples to padd
thresh_track_s=[6,300] # time-constants for tracking slow drift in offset
def trackDroppedSamples(nsamp,ts_ms):
    '''function to detect dropped samples so can insert padding'''
    global nerr, estndropped, samp0
    if nsamp<fSample*10 : # warm-up the offset
        samp0est = nsamp - ts_ms*fSample/1000
        samp0 = max(samp0,samp0est) if not samp0 is None else samp0est
        estndropped=fSample*10
        return 0

    # update the tracking info
    samp_est = int(ts_ms/1000*fSample + samp0)
    err = samp_est - nsamp
    # slow tracking of sample offset
    if err < 0 : # proportional step for neg err => quadratic-lowerbound
        samp0 = samp0 - err/(thresh_track_s[0]*fSample)
    else : # fixed step for positive err => linear upper-bound
        samp0 = samp0 - min(1,err)/(thresh_track_s[1]*fSample)

    if err < epsilon_err :
        # inside deadzone
        nerr = 0
        return 0

    # outside the deadzone, i.e. err>=epsilon_err
    # track the total time with consistently fewer samples than expected
    nerr = nerr + 1
    
    if nerr>thresh_nerr_s*fSample/2 : # record the size of the offset
        estndropped = min(estndropped,err)

    # if long enough with missing samples -> mark to insert the missing
    if nerr>thresh_nerr_s*fSample :
        nerr    = 0
        ndropped= estndropped
        estndropped = fSample*10 # start with inf
        return ndropped # return the delta to the error



def processandpadsample(sample):
    global nSamp, ftBlock, lastdata, lastsamp, lastcount
    data=[d*SCALE_FACTOR_EEG for d in sample.channels_data] # [ nEEG ]
    sample.sample_number=sample.id

    # get the sample number we are looking for
    if lastsamp :
      nextsamp,nextcount=incsampCount(lastsamp,lastcount)
    else:
      nextsamp =sample.sample_number
      nextcount=0 # always start with 1st sample in packet

    if len(data)==0 : 
        print("%d) Packet with no channel data, skipped!"%(nSamp))
        return
    
    # check for missing samples, and pad if necessary   
    if not sample.sample_number==nextsamp :
        print("%d) skipped a sample! (%d,%d)->%d, padding: "%(nSamp,lastsamp,lastcount,sample.sample_number),end='')
        if not lastdata is None :
            padsamp=0
            while not nextsamp == sample.sample_number:
                ftBlockAppend(lastdata,nextsamp,nextcount,'s')
                nextsamp,nextcount = incsampCount(nextsamp,nextcount)
                print("s",end="")
            print()

    # second check based on tracking the sample arrival time
    ndroppedsamp = trackDroppedSamples(nSamp,getTime_ms())
    if ndroppedsamp :
        print("%d) %d dropped samples detected, padding: "%(nSamp,ndroppedsamp))
        if not lastdata is None :
            for i in range(ndroppedsamp) : 
                ftBlockAppend(lastdata,lastsamp,lastcount,'d')
                print("d",end="")
            print()
        print("%d %d: (%d,%d)->%d end-dropped"%(getTime_ms(),nSamp,lastsamp,lastcount,sample.sample_number))
      
    # finally put the actual new data
    # put this data sample and updated tracking info
    ftBlockAppend(data,sample.sample_number,None)


def ftBlockAppend(data,sample_number,sample_count=None,why=0):
    '''function to add a new data sample to the block data to send to ftbuffer,
        and log info about where the data came from'''
    global nSamp, ftBlock, lastdata, lastsamp, lastcount
    ftBlock.append(data)
    nSamp = nSamp + 1
    # store info on the last data we have processed
    lastdata = data
    if sample_count is None:
        lastcount = 1 if sample_number==lastsamp else 0
    else:
        lastcount = sample_count
    lastsamp  = sample_number
    if logFile :
        why = ord(why) if type(why) is str else why
        print("%d %d %d %d %d"%(getTime_ms(),nSamp,why,lastsamp,lastcount),file=logFile)
    return nSamp
    
def utopia_putsamples(samples):
    '''process multiple samples at a time'''
    global nSamp, nBlock, ftBlock
    for sample in samples:
        processandpadsample(sample)

    if len(ftBlock)>ftBlockSample:
        client.getNewMessages(0) # flush the incomming message queue
        client.sendMessage(utopiaclient.DataPacket(getTime_ms(),ftBlock))
        nBlock = nBlock+1
        ftBlock= []
    printLog(nSamp,nBlock)    

def printLog(nSamp,nBlock):
    ''' textual logging of the data arrivals etc.'''
    global t0, nextLogTime
    if getTime() > nextLogTime:
        elapsed = getTime()-t0
        print("%d %d %f %f (samp,blk,s,hz)"%(nSamp,nBlock,elapsed,nSamp/elapsed),flush=True)
        nextLogTime=getTime()+LOGINTERVAL_S
    
board=None
client=None
def initConnections(host=None,obcimac=None):
    global board
    global client
    global fSample
    print("Opening OpenBCI Ganglion")
    board = OpenBCIGanglion(mac=obcimac,sample_rate=200)
    nChans = 4
    fSample= board.sample_rate
    print("Opened openBCI %s on %s with %d ch @ %g hz"%(board.board_type,obcimac,nChans,fSample))
    print("Opening ftBuffer connection") 

    client=utopiaclient.UtopiaClient()
    client.disableHeartbeats() # disabled for data sources
    client.autoconnect(host)
    # don't subscribe to anything
    client.sendMessage(utopiaclient.Subscribe(getTime_ms(),""));
    print("Putting header.")    
    client.sendMessage(utopiaclient.DataHeader(getTime_ms(),nChans, fSample, ""))
    return (board,client)

def readConfigFile(fname):
    '''Read the ftbuffer_ganglion cofiguration from a file'''
    config={}
    with open(fname,'r') as f:
        # parse into a dict
        for line in f:
            if line.startswith("#") : continue
            if "=" in line : 
                key,val=line.split("=",1)
                config[key.strip()]=val.strip()
    return config

def run(host=None, obcimac=None):    
    print("ftbuffer hostname:  %s"%(host))
    print("Ganglion MACadress: %s"%(obcimac))
    initConnections(host=host,obcimac=obcimac)
    # record the stream start time
    global t0, nextLogTime
    t0=getTime()
    nextLogTime=t0
    
    global logFile
    if logFile : 
        logFile = open(logFile,"w")
    
    # start the forwarder
    print("Starting the data forwarder")
    board.start_stream(utopia_putsamples)


def parse_arguments():
    import sys
    argv = sys.argv
    print("Arguments:",*argv)
    # N.B. argv0 is the script name
    if len(argv)>2 :
        host=argv[1] if len(argv)>1 else None
        obcimac=argv[2] if len(argv)>2 else None
    elif len(argv)==2 : # single argument, read from config file
        try:
            print("Reading config from file : {}".format(argv[1]))
            config=readConfigFile(argv[1])
            obcimac=config['ganglionmac'] if 'ganglionmac' in config else None
        except FileNotFoundError :
            host=argv[2]
    return dict(host=host, obcimac=obcimac)

if __name__=="__main__":
    args = parse_arguments()
    run(*args)



import numpy as np
from mindaffectBCI.decoder.utils import RingBuffer

class lower_bound_tracker():
    """ sliding window linear trend tracker
    """   

    def __init__(self, window_size=200, C=(.01,None), outlier_thresh=(1,None), step_size=10, step_threshold=2, a0=1, b0=0, warmup_size=10):
        self.window_size = window_size
        self.step_size = int(step_size*window_size) if step_size<1 else step_size
        self.a0 = a0 if a0 is not None else 1
        self.b0 = b0
        self.warmup_size = int(warmup_size*window_size) if warmup_size<1 else warmup_size
        self.step_threshold = step_threshold
        self.outlier_thresh = outlier_thresh
        if not hasattr(self.outlier_thresh,'__iter__'):
            self.outlier_thresh = (self.outlier_thresh, self.outlier_thresh)
        self.C = C
        if not hasattr(self.C,'__iter__'):
            self.C = None

    def reset(self, keep_model=False):
        self.buffer.clear()
        if not keep_model:
            self.a = self.a0
            self.b = self.b0

    def fit(self,X,Y):
        self.buffer = RingBuffer(self.window_size, (2,), dtype=X.dtype if isinstance(X,np.ndarray) else float)
        self.append(X,Y)
        self.a = self.a0
        self.b = self.buffer[-1,1] - self.a * self.buffer[-1,0]

    def append(self,X,Y):
        if isinstance(X, np.ndarray) and X.ndim > 1:
            self.buffer.extend(np.hstack((X,Y)))
        else:
            self.buffer.append(np.array((X,Y)))

    def transform(self,X,Y):
        if not hasattr(self,'buffer'):
            self.fit(X,Y)
            return Y
        
        # add into ring-buffer
        self.append(X,Y)

        # update the fit
        self.update()

        # return Yest
        return self.getY(X)

    def update(self):
        if self.buffer.shape[0]>0 and self.buffer.shape[0] < self.warmup_size :
            if self.buffer.shape[0]>2:
                self.a = np.median(np.diff(self.buffer[:,1])/np.diff(self.buffer[:,0]))
            self.b = self.buffer[-1,1] - self.a * self.buffer[-1,0]
            return

        # get the data
        x = self.buffer[:,0].astype(float)
        y = self.buffer[:,1].astype(float)

        # center the data to improve the numerical robustness
        mu_x  = np.median(x)
        x = x - mu_x
        mu_y = np.median(y)
        y = y - mu_y

        # IRWLS fit to the data, with outlier suppression and asymetric huber loss
        # add constant feature  for the intercept
        x = np.append(x[:,np.newaxis],np.ones((x.shape[0],1),dtype=x.dtype),1)
        # weighted LS  solve
        sqrt_w = np.ones((x.shape[0],),dtype=x.dtype) # square root of weight so can use linalg.lstsq
        # start from previous solution
        a = self.a
        b = self.b + self.a*mu_x - mu_y # shift bias to compensate for shift in x/y
        # TODO[]: convergence criteria, to stop early
        for i in range(3):
            y_est = x[:,0]*a + b
            err = y - y_est # server > true, clip positive errors
            mu_err = np.mean(err)
            #print('{} mu_err={}\n'.format(i,mu_err))
            scale = np.mean(np.abs(err-mu_err))

            # 1 by default
            sqrt_w[:] = 1.0

            if self.C is not None:
                # sign based cost function

                # postive errors
                if self.C[0] is not None:
                    pos_l1 = err > self.C[0]
                    # wght = C/abs(err)
                    sqrt_w[pos_l1] = np.sqrt(self.C[0]/err[pos_l1])

                if self.C[1] is not None:
                    neg_l1 = err < -self.C[1]
                    sqrt_w[neg_l1] = np.sqrt(self.C[1]/-err[neg_l1])

            if self.outlier_thresh is not None:

                if self.outlier_thresh[0] is not None:
                    # clip over-estimates
                    pos_outlier = err-mu_err > self.outlier_thresh[0]*scale
                    sqrt_w[pos_outlier] = 0
                    #print("{} overestimates".format(np.sum(clipIdx)))
                    #y_fit[pos_outlier] = y_est[pos_outlier] + self.outlier_thresh[0]*scale

                if self.outlier_thresh[1] is not None:
                    # clip under-estimates
                    neg_outlier = err-mu_err < -self.outlier_thresh[1]*scale
                    sqrt_w[neg_outlier] = 0
                    #print("{} underestimates".format(np.sum(clipIdx)))
                    #y_fit[neg_outlier] = y_est[neg_outlier] - self.outlier_thresh[1]*scale

            # solve weighted least squares
            ab, _, _, _ = np.linalg.lstsq(x*sqrt_w[:,np.newaxis], y*sqrt_w, rcond=-1)
            # extract parameters
            a = ab[0]
            b = ab[1]
                    
        self.a = a
        self.b = b - self.a*mu_x + mu_y 

        # check for steps in the last part of the data - if enough data
        if err.shape[0]> 2*self.step_size:
            mu_err = np.median(err[:-self.step_size])
            std_err = np.std(err[:-self.step_size])
            mu = np.median(err[-self.step_size:])
            if mu > mu_err + self.step_threshold * std_err or mu < mu_err - self.step_threshold * std_err:
                #print('step detected!')
                self.reset(keep_model=True)

    def getX(self,y):
        return ( y  - self.b ) / self.a 

    def getY(self,x):
        return self.a * x + self.b

    @staticmethod
    def testcase(savefile='-'):
        if savefile is None:
            np.random.seed(0) # make reproducable
            X = np.arange(1000,dtype=np.float64) + np.random.randn(1)*1e6
            a = np.random.uniform(0,1)*10+10 
            b = 1000*np.random.randn(1) +.3
            Ytrue= a*X+b
            # add some steps
            for i in range(5):
                stepi = np.random.randint(Ytrue.shape[0])
                Ytrue[stepi:] += np.random.randint(3)*a
            N = np.random.standard_normal(Ytrue.shape)
            #Y    = Ytrue+ *10
            Y = Ytrue + np.exp((N-1)*3)
            print("{}) a={} b={}".format('true',a,b))

        else: # load from file
            import glob
            import os
            files = glob.glob(os.path.join(os.path.dirname(os.path.abspath(__file__)),'../../logs/mindaffectBCI*.txt')) # * means all if need specific format then *.csv
            savefile = max(files, key=os.path.getctime)
            #savefile = "C:\\Users\\Developer\\Downloads\\mark\\mindaffectBCI_brainflow_200911_1339.txt" 
            #savefile = "C:/Users/Developer/Downloads/khash/mindaffectBCI_brainflow_ipad_200908_1938.txt"
            from mindaffectBCI.decoder.offline.read_mindaffectBCI import read_mindaffectBCI_messages
            from mindaffectBCI.utopiaclient import DataPacket
            print("Loading: {}".format(savefile))
            msgs = read_mindaffectBCI_messages(savefile,regress=None) # load without time-stamp fixing.
            dp = [ m for m in msgs if isinstance(m,DataPacket)]
            nsc = np.array([ (m.samples.shape[0],m.sts,m.timestamp) for m in dp])
            X = np.cumsum(nsc[:,0])
            Y = nsc[:,1]
            Ytrue = nsc[:,2]

        #ltt = lower_bound_tracker(window_size=100, outlier_thresh=(.5,None), step_size=10, step_threshold=2)#, C=(.1,None))
        # differential loss.  Linear above, quadratic below
        #ltt = lower_bound_tracker(window_size=200, C=(.1,None), step_size=10, step_threshold=2)#, C=(.1,None))
        # both, lower-bound + clipping
        ltt = lower_bound_tracker(window_size=100, outlier_thresh=(1,None), C=(.1,None), step_size=10, step_threshold=2)
        ltt.fit(X[0],Y[0]) # check scalar inputs
        step = 1
        idxs = list(range(1,X.shape[0],step))
        ab = np.zeros((len(idxs),2))
        dts = np.zeros((Y.shape[0],))
        dts[0] = ltt.getY(X[0])
        for i,j in enumerate(idxs):
            dts[j:j+step] = ltt.transform(X[j:j+step],Y[j:j+step])
            ab[i,:] = (ltt.a,ltt.b)
            yest = ltt.getY(X[j])
            err =  yest - Y[j]
            if abs(err)> 1000:
                print("{}) argh! yest={} ytrue={} err={}".format(i,yest,Ytrue[j],err))
            if i < 100:
                print("{:4d}) a={:5f} b={:5f}\ty_est-y={:2.5f}".format(j,ab[i,0],ab[i,1],
                        Y[j]-yest))

        import matplotlib.pyplot as plt
        ab,res,_,_ = np.linalg.lstsq(np.append(X[:,np.newaxis],np.ones((X.shape[0],1)),1),Y,rcond=-1)
        ots = X*ab[0]+ab[1]        
        idx=range(X.shape[0])
        plt.plot((Y[idx]-Y[0])- (X[idx]-X[0])*ab[0],'.-',label='server ts')
        plt.plot((dts[idx]-Y[0]) - (X[idx]-X[0])*ab[0],'.-',label='regressed ts (samp vs server)')
        plt.plot((ots[idx]-Y[0]) - (X[idx]-X[0])*ab[0],'.-',label='regressed ts (samp vs server) offline')

        err = (Y-Y[0]) - (X-X[0])*ab[0]
        cent = np.median(err); scale=np.median(np.abs(err-cent))
        plt.ylim((cent-scale*5,cent+scale*5))
        plt.grid()
        plt.legend()
        plt.show()

if __name__ == "__main__":
    lower_bound_tracker.testcase('-')


#!/usr/bin/env python3

# Copyright (c) 2019 MindAffect B.V.
#  Author: Jason Farquhar <jason@mindaffect.nl>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


# get the general noisetagging framework
from mindaffectBCI.noisetag import Noisetag
from mindaffectBCI.utopiaclient import DataPacket

# graphic library
import pyglet
isi = 1/60
drawrate = 0  # rate at which draw is called

class Screen:

    '''Screen abstract-class which draws stuff on the screen until finished'''
    def __init__(self, window):
        self.window = window

    def reset(self):
        '''reset this screen to clean state'''
        pass

    def draw(self, t):
        '''draw the display, N.B. NOT including flip!'''
        pass

    def is_done(self):
        '''test if this screen wants to quit'''
        return False






#-----------------------------------------------------------------
#-----------------------------------------------------------------
#-----------------------------------------------------------------
#-----------------------------------------------------------------
class InstructionScreen(Screen):
    '''Screen which shows a textual instruction for duration or until key-pressed'''
    def __init__(self, window, text, duration=5000, waitKey=True):
        super().__init__(window)
        self.t0 = None # timer for the duration
        self.duration = duration
        self.waitKey = waitKey
        self.isRunning = False
        self.isDone = False
        self.clearScreen = True
        # initialize the instructions screen
        self.instructLabel = pyglet.text.Label(x=window.width//2,
                                             y=window.height//2,
                                             anchor_x='center',
                                             anchor_y='center',
                                             font_size=24,
                                             color=(255, 255, 255, 255),
                                             multiline=True,
                                             width=int(window.width*.8))
        self.set_text(text)
        print("Instruct (%dms): %s"%(duration, text))

    def reset(self):
        self.isRunning = False
        self.isDone = False

    def set_text(self, text):
        '''set/update the text to show in the instruction screen'''
        if type(text) is list:
            text = "\n".join(text)
        self.instructLabel.begin_update()
        self.instructLabel.text=text
        self.instructLabel.end_update()

    def is_done(self):
        # check termination conditions
        if not self.isRunning:
            self.isDone = False
            return self.isDone
        if self.waitKey:
            #global last_key_press
            if self.window.last_key_press:
                self.key_press = self.window.last_key_press
                self.isDone = True
                self.window.last_key_press = None
        if self.elapsed_ms() > self.duration:
            self.isDone = True
        return self.isDone

    def elapsed_ms(self):
        return getTimeStamp()-self.t0

    def draw(self, t):
        '''Show a block of text to the user for a given duration on a blank screen'''
        if not self.isRunning:
            self.isRunning = True  # mark that we're running
            self.t0 = getTimeStamp()
        if self.clearScreen:
            self.window.clear()
        self.instructLabel.draw()





#-----------------------------------------------------------------
#-----------------------------------------------------------------
#-----------------------------------------------------------------
#-----------------------------------------------------------------
#-----------------------------------------------------------------
class MenuScreen(InstructionScreen):
    '''Screen which shows a textual instruction for duration or until key-pressed'''
    def __init__(self, window, text, valid_keys):
        super().__init__(window, text, 99999999, True)
        self.menu_text = text
        self.valid_keys = valid_keys
        self.key_press = None
        print("Menu")

    def set_message(self,message:str):
        self.set_text(self.menu_text+'\n\n\n'+message)

    def is_done(self):
        # check termination conditions
        if not self.isRunning:
            self.isDone = False
            return self.isDone

        # check if should update display
        # TODO[]: only update screen 1x / second
        global flipstats
        flipstats.update_statistics()
        self.set_message("Frame-duration: {:4.1f} +/-{:4.1f}ms".format(flipstats.med,flipstats.sigma))

        global last_key_press
        if self.window.last_key_press:
            self.key_press = self.window.last_key_press
            if self.key_press in self.valid_keys:
                self.isDone = True
            self.window.last_key_press = None
        if self.elapsed_ms() > self.duration:
            self.isDone = True
        return self.isDone





#-----------------------------------------------------------------
#-----------------------------------------------------------------
#-----------------------------------------------------------------
#-----------------------------------------------------------------
#-----------------------------------------------------------------
class ResultsScreen(InstructionScreen):
    '''Modified instruction screen with waits for and presents calibration results'''

    waiting_text = "Waiting for performance results from decoder\n\nPlease wait"
    results_text = "Calibration Performance: %3.0f%% Correct\n\nKey to continue"
    def __init__(self, window, noisetag, duration=20000, waitKey=False):
        super().__init__(window, self.waiting_text, duration, waitKey)
        self.noisetag = noisetag
        self.pred = None

    def reset(self):
        self.noisetag.clearLastPrediction()
        self.pred = None
        super().reset()

    def draw(self, t):
        '''check for results from decoder.  show if found..'''
        if not self.isRunning:
            self.reset()
        # check for new predictions
        pred = self.noisetag.getLastPrediction()
        # update text if got predicted performance
        if pred is not None and (self.pred is None or pred.timestamp > self.pred.timestamp) :
            self.pred = pred
            print("Prediction:{}".format(self.pred))
            self.waitKey = True
            self.set_text(self.results_text%((1.0-self.pred.Perr)*100.0))
        super().draw(t)






#-----------------------------------------------------------------
#-----------------------------------------------------------------
#-----------------------------------------------------------------
#-----------------------------------------------------------------
#-----------------------------------------------------------------
class ConnectingScreen(InstructionScreen):
    '''Modified instruction screen with waits for the noisetag to connect to the decoder'''

    prefix_text = "Welcome to the mindaffectBCI\n\n"
    searching_text = "Searching for the mindaffect decoder\n\nPlease wait"
    trying_text = "Trying to connect to: %s\n Please wait"
    connected_text = "Success!\nconnected to: %s"
    query_text = "Couldnt auto-discover mindaffect decoder\n\nPlease enter decoder address: %s"
    drawconnect_timeout_ms = 50
    autoconnect_timeout_ms = 5000

    def __init__(self, window, noisetag, duration=150000):
        super().__init__(window, self.prefix_text + self.searching_text, duration, False)
        self.noisetag = noisetag
        self.host = None
        self.port = -1
        self.usertext = ''
        self.stage = 0

    def draw(self, t):
        '''check for results from decoder.  show if found..'''
        global last_text, last_key_press
        if not self.isRunning:
            super().draw(t)
            return

        if not self.noisetag.isConnected():
            if self.stage == 0: # try-connection
                print('Not connected yet!!')
                self.noisetag.connect(self.host, self.port,
                                      queryifhostnotfound=False,
                                      timeout_ms=self.drawconnect_timeout_ms)
                if self.noisetag.isConnected():
                    self.set_text(self.prefix_text + self.connected_text%(self.noisetag.gethostport()))
                    self.t0 = getTimeStamp()
                    self.duration = 1000
                    self.noisetag.subscribe("MSPQ")
                elif self.elapsed_ms() > self.autoconnect_timeout_ms:
                    # waited too long, giveup and ask user
                    self.stage = 1
                    # ensure old key-presses are gone
                    self.window.last_text = None
                    self.window.last_key_press = None

            elif self.stage == 1:  # query hostname
                # query the user for host/port
                # accumulate user inputs
                if self.window.last_key_press:
                    if self.window.last_key_press == pyglet.window.key.BACKSPACE:
                        # remove last character
                        self.usertext = self.usertext[:-1]
                    self.window.last_key_press = None
                if self.window.last_text:
                    print(self.window.last_text + ":" + str(ord(self.window.last_text)))
                    if self.window.last_text == '\n' or self.window.last_text == '\r':
                        # set as new host to try
                        self.host = self.usertext
                        self.usertext = ''
                        self.set_text(self.prefix_text + self.trying_text%(self.host))
                        self.stage = 0 # back to try-connection stage
                    elif self.window.last_text:
                        # add to the host string
                        self.usertext += last_text
                    self.window.last_text = None
                if self.stage == 1: # in same stage
                    # update display with user input
                    self.set_text(self.prefix_text + self.query_text%(self.usertext))
        super().draw(t)






#-----------------------------------------------------------------
#-----------------------------------------------------------------
#-----------------------------------------------------------------
#-----------------------------------------------------------------
#-----------------------------------------------------------------
class SettingsScreen(InstructionScreen):
    '''Modified instruction screen to change various settings - selection threshold'''

    prefix_text = "Configuration Settings\n\n"
    threshold_text = "New Selection Threshold: %s\n"

    def __init__(self, window, settings_class, duration=150000):
        super().__init__(window, self.prefix_text + self.threshold_text%(settings_class.selectionThreshold), duration, False)
        self.settings_class = settings_class
        self.usertext = ''

    def draw(self, t):
        '''check for results from decoder.  show if found..'''
        global last_text, last_key_press
        if not self.isRunning:
            super().draw(t)
            return

        # query the user for host/port
        # accumulate user inputs
        if self.window.last_key_press:
            if self.window.last_key_press == pyglet.window.key.BACKSPACE:
                # remove last character
                self.usertext = self.usertext[:-1]
            self.window.last_key_press = None
            if self.window.last_text:
                print(self.window.last_text + ":" + str(ord(self.window.last_text)))
            if self.window.last_text == '\n' or self.window.last_text == '\r':
                # set as new host to try
                try:
                    self.threshold = float(self.usertext)
                    self.settings_class.selectionThreshold=self.threshold
                    self.isDone = True
                except ValueError:
                    # todo: flash to indicate invalid..
                    pass
            elif self.window.last_text and self.window.last_text in "0123456789.":
                # add to the host string
                self.usertext += window.last_text
            self.window.last_text = None
            self.set_text(self.prefix_text + self.threshold_text%(self.usertext))
        super().draw(t)


#-----------------------------------------------------------------
class QueryDialogScreen(InstructionScreen):
    '''Modified instruction screen queries the user for textual input'''

    def __init__(self, window, text, duration=50000, waitKey=True):
        super().__init__(window, text, 50000, False)
        self.query = text
        self.usertext = ''

    def draw(self, t):
        '''check for results from decoder.  show if found..'''
        # query the user for host/port
        # accumulate user inputs
        global last_key_press, last_text

        if self.window.last_key_press:
            if self.window.last_key_press == pyglet.window.key.BACKSPACE:
                self.usertext = self.usertext[:-1]
                self.set_text(self.query +self.usertext)
            self.window.last_key_press = None
        if self.window.last_text:
            if self.window.last_text == '\r' or self.window.last_text == '\n':
                self.isDone = True
            elif self.window.last_text:
                # add to the host string
                self.usertext += self.window.last_text
            self.window.last_text=None
            # update display with user input
            self.set_text(self.query +self.usertext)
        super().draw(t)




#-----------------------------------------------------------------
#-----------------------------------------------------------------
#-----------------------------------------------------------------
#-----------------------------------------------------------------
from math import log10
from collections import deque
class ElectrodequalityScreen(Screen):
    '''Screen which shows the electrode signal quality information'''

    instruct = "Electrode Quality\n\nAdjust headset until all electrodes are green\n(or noise to signal ratio < 5)"
    def __init__(self, window, noisetag, nch=4, duration=3600*1000, waitKey=True):
        super().__init__(window)
        self.noisetag = noisetag
        self.t0 = None  # timer for the duration
        self.duration = duration
        self.waitKey = waitKey
        self.clearScreen = True
        self.isRunning = False
        self.update_nch(nch)
        self.dataringbuffer = deque()  # deque so efficient sliding data window
        self.datawindow_ms = 4000  # 5seconds data plotted
        self.datascale_uv = 20  # scale of gap between ch plots
        print("Electrode Quality (%dms)"%(duration))

    def update_nch(self, nch):
        self.batch      = pyglet.graphics.Batch()
        self.background = pyglet.graphics.OrderedGroup(0)
        self.foreground = pyglet.graphics.OrderedGroup(1)
        winw, winh = window.get_size()
        r = (winh*.8)/(nch+1)
        # TODO[X] use bounding box
        self.chrect = (int(winw*.1), 0, r, r) # bbox for each signal, (x, y, w, h)
        # make a sprite to draw the electrode qualities
        img = pyglet.image.SolidColorImagePattern(color=(255, 255, 255, 255)).create_image(2, 2)
        # anchor in the center to make drawing easier
        img.anchor_x = 1
        img.anchor_y = 1
        self.sprite = [None]*nch
        self.label  = [None]*nch
        self.linebbox = [None]*nch # bounding box for the channel line
        for i in range(nch):
            x = self.chrect[0]
            y = self.chrect[1]+(i+1)*self.chrect[3]
            # convert to a sprite and make the right size
            self.sprite[i] = pyglet.sprite.Sprite(img, x=x, y=y,
                                                batch=self.batch,
                                                group=self.background)
            # make the desired size
            self.sprite[i].update(scale_x=r*.6/img.width, scale_y=r*.6/img.height)
            # and a text label object
            self.label[i] = pyglet.text.Label("%d"%(i), font_size=32,
                                            x=x, y=y,
                                            color=(255, 255, 255, 255),
                                            anchor_x='center',
                                            anchor_y='center',
                                            batch=self.batch,
                                            group=self.foreground)
            # bounding box for the datalines
            self.linebbox[i] = (x+r, y, winw-(x+r)-.5*r, self.chrect[3])
        # title for the screen
        self.title=pyglet.text.Label(self.instruct, font_size=32,
                                     x=winw*.1, y=winh, color=(255, 255, 255, 255),
                                     anchor_y="top",
                                     width=int(winw*.9),
                                     multiline=True,
                                     batch=self.batch,
                                     group=self.foreground)

    def reset(self):
        self.isRunning = False

    def is_done(self):
        # check termination conditions
        isDone=False
        if not self.isRunning:
            return False
        if self.waitKey:
            global last_key_press
            if self.window.last_key_press:
                self.key_press = self.window.last_key_press
                isDone = True
                self.window.last_key_press = None
        if getTimeStamp() > self.t0+self.duration:
            isDone=True
        if isDone:
            self.noisetag.removeSubscription("D")
            self.noisetag.modeChange("idle")
        return isDone

    def draw(self, t):
        '''Show a set of colored circles based on the lastSigQuality'''
        if not self.isRunning:
            self.isRunning = True # mark that we're running
            self.t0 = getTimeStamp()
            self.noisetag.addSubscription("D") # subscribe to "DataPacket" messages
            self.noisetag.modeChange("ElectrodeQuality")
            self.dataringbuffer.clear()
        if self.clearScreen:
            self.window.clear()
        # get the sig qualities
        electrodeQualities = self.noisetag.getLastSignalQuality()
        if not electrodeQualities: # default to 4 off qualities
            electrodeQualities = [.5]*len(self.sprite)

        if len(electrodeQualities) != len(self.sprite):
            self.update_nch(len(electrodeQualities))

        issig2noise = True #any([s>1.5 for s in electrodeQualities])
        # update the colors
        #print("Qual:", end='')
        for i, qual in enumerate(electrodeQualities):
            self.label[i].text = "%d: %3.1f"%(i+1, qual)
            #print(self.label[i].text + " ", end='')
            if issig2noise:
                qual = log10(qual)/1 # n2s=50->1 n2s=10->.5 n2s=1->0
            qual = max(0, min(1, qual))
            qualcolor = (int(255*qual), int(255*(1-qual)), 0) #red=bad, green=good
            self.sprite[i].color=qualcolor
        #print("")
        # draw the updated batch
        self.batch.draw()

        # get the raw signals
        msgs=self.noisetag.getNewMessages()
        for m in msgs:
            if m.msgID == DataPacket.msgID:
                print('D', end='', flush=True)
                self.dataringbuffer.extend(m.samples)
                if getTimeStamp() > self.t0+self.datawindow_ms: # slide buffer
                    # remove same number of samples we've just added
                    for i in range(len(m.samples)):
                        self.dataringbuffer.popleft()


        if self.dataringbuffer:
            if len(self.dataringbuffer[0]) != len(self.sprite):
                self.update_nch(len(self.dataringbuffer[0]))

            # transpose and flatten the data
            # and estimate it's summary statistics
            from statistics import median

            # CAR
            dataringbuffer =[]
            for t in self.dataringbuffer:
                mu = median(t)
                dataringbuffer.append([c-mu for c in t])
            
            # other pre-processing
            data = []
            mu = [] # mean
            mad = [] # mean-absolute-difference
            nch=len(self.linebbox)
            for ci in range(nch):
                d = [ t[ci] for t in dataringbuffer ]
                # mean last samples
                tmp = d[-int(len(d)*.2):]
                mui = sum(tmp)/len(tmp)
                # center (in time)
                d = [ t-mui for t in d ]
                # scale estimate
                madi = sum([abs(t-mui) for t in tmp])/len(tmp)
                data.append(d)
                mu.append(mui)
                mad.append(madi)

            
            datascale_uv = max(5,median(mad)*4)

            for ci in range(nch):
                d = data[ci]
                # map to screen coordinates
                bbox=self.linebbox[ci]

                # downsample if needed to avoid visual aliasing
                #if len(d) > (bbox[2]-bbox[1])*2:
                #    subsampratio = int(len(d)//(bbox[2]-bbox[1]))
                #    d = [d[i] for i in range(0,len(d),subsampratio)]

                # map to screen coordinates
                xscale = bbox[2]/len(d)
                yscale = bbox[3]/datascale_uv #datascale_uv # 10 uV between lines
                y = [ bbox[1] + s*yscale for s in d ]
                x = [ bbox[0] + i*xscale for i in range(len(d)) ]
                # interleave x, y to make gl happy
                coords = tuple( c for xy in zip(x, y) for c in xy )
                # draw this line
                col = [0,0,0]; col[ci%3]=1
                pyglet.graphics.glColor3d(*col)
                pyglet.gl.glLineWidth(1)
                pyglet.graphics.draw(len(d), pyglet.gl.GL_LINE_STRIP, ('v2f', coords))

                # axes scale
                x = bbox[0]+bbox[2]+20 # at *right* side of the line box
                y = bbox[1]
                pyglet.graphics.glColor3f(1,1,1)
                pyglet.gl.glLineWidth(10)
                pyglet.graphics.draw(2, pyglet.gl.GL_LINES,
                                        ('v2f', (x,y-10/2*yscale, x,y+10/2*yscale)))





#-----------------------------------------------------------------
#-----------------------------------------------------------------
#-----------------------------------------------------------------
#-------------------------------------------------------------
class FrameRateTestScreen(InstructionScreen):
    ''' screen from testing the frame rate of the display under pyglet control '''

    testing_text = "Checking display framerate\nPlease wait"
    results_text = "Frame rate: "
    failure_text = "WARNING:\nhigh variability in frame timing detected\nyour performance may suffer\n"
    success_text = "SUCCESS:\nsufficient accuracy frame timing detected\n"
    statistics_text = "\n{:3.0f} +/-{:3.1f} [{:2.0f},{:2.0f}]\n mean +/-std [min,max]"
    closing_text = "\n Press key to continue."
    
    def __init__(self, window, testduration=2000, warmup_duration=1000, duration=20000, waitKey=False):
        super().__init__(window, self.testing_text, duration, waitKey)
        self.testduration = testduration
        self.warmup_duration = warmup_duration
        self.ftimes = []
        self.logtime = None
        self.log_interval = 2000
        
    def draw(self, t):
        if not self.isRunning:
            self.ftimes = []
            self.logtime = 0
        # call parent draw method
        super().draw(t)
                    
        # record the  flip timing info
        # TODO[]: use a deque to make sliding window...
        # TODO[]: plot the histogram of frame-times?
        if self.elapsed_ms() > self.warmup_duration:
            self.ftimes.append(self.window.lastfliptime)

        if self.elapsed_ms() > self.warmup_duration + self.testduration:
            if self.elapsed_ms() > self.logtime:
                self.logtime=self.elapsed_ms() + self.log_interval
                log=True
            else:
                log=False
            (medt,madt,mint,maxt) = self.analyse_ftimes(self.ftimes,log)
            # show warning if timing is too poor
            if madt > 1:
                msg=self.failure_text
            else:
                msg=self.success_text
            msg += self.statistics_text.format(medt,madt,mint,maxt)
            msg += self.closing_text
            self.set_text(msg)
            self.waitKey = True

    @staticmethod
    def analyse_ftimes(ftimes, verb=0):
        # convert to inter-frame time
        fdur = [ ftimes[i+1]-ftimes[i] for i in range(len(ftimes)-1) ]
        #print(["%d"%(int(f)) for f in fdur])
        # analyse the frame durations, in outlier robust way
        from statistics import median    
        medt=median(fdur) # median (mode?)
        madt=0; mint=999; maxt=-999; N=0;
        for dt in fdur:
            if dt > 200 : continue # skip outliers
            N=N+1
            madt += (dt-medt) if dt>medt else (medt-dt)
            mint = dt if dt<mint else mint
            maxt = dt if dt>maxt else maxt
        madt = madt/len(fdur)

        if verb>0 :
            print("Statistics: %f(%f) [%f,%f]"%(medt,madt,mint,maxt))
            try:    
                from numpy import histogram
                [hist,bins]=histogram(fdur,range(8,34,2))
                # report summary statistics to the user
                print("Histogram:",
                      "\nDuration:","\t".join("%6.4f"%((bins[i]+bins[i+1])/2) for i in range(len(bins)-1)),
                      "\nCount   :","\t".join("%6d"%t for t in hist))
            except:
                pass
        return (medt,madt,mint,maxt)





#-----------------------------------------------------------------
#-----------------------------------------------------------------
#-----------------------------------------------------------------
#-------------------------------------------------------------
class SelectionGridScreen(Screen):
    '''Screen which shows a grid of symbols which will be flickered with the noisecode
    and which can be selected from by the mindaffect decoder Brain Computer Interface'''

    LOGLEVEL=0

    def __init__(self, window, symbols, noisetag, objIDs=None,
                 bgFraction=.2, instruct="", 
                 clearScreen=True, sendEvents=True, liveFeedback=True, optosensor=True,
                 waitKey=True, stimulus_callback=None):
        '''Intialize the stimulus display with the grid of strings in the
        shape given by symbols.
        Store the grid object in the fakepresentation.objects list so can
        use directly with the fakepresentation BCI presentation wrapper.'''
        self.window=window
        # create set of sprites and add to render batch
        self.clearScreen= True
        self.isRunning=False
        self.isDone=False
        self.sendEvents=sendEvents
        self.liveFeedback=liveFeedback
        self.framestart = getTimeStamp()
        self.frameend = getTimeStamp()
        self.symbols = symbols
        self.objIDs = objIDs
        self.optosensor = optosensor
        # N.B. noisetag does the whole stimulus sequence
        self.set_noisetag(noisetag)
        self.set_grid(symbols, objIDs, bgFraction, sentence=instruct)
        self.liveSelections = None
        self.feedbackThreshold = 1
        self.waitKey=waitKey
        self.stimulus_callback = stimulus_callback

    def reset(self):
        self.isRunning=False
        self.isDone=False
        self.nframe=0
        self.set_grid()

    def set_noisetag(self, noisetag):
        self.noisetag=noisetag

    def setliveFeedback(self, value):
        self.liveFeedback=value

    def setliveSelections(self, value):
        if self.liveSelections is None :
            self.noisetag.addSelectionHandler(self.doSelection)
        self.liveSelections = value

    def getSymb(self,idx):
        ii=0
        for i in range(len(self.symbols)): # rows
            for j in range(len(self.symbols[i])): # cols
                if self.symbols[i][j] is None: continue
                if ii==idx:
                    return self.symbols[i][j]
                ii = ii + 1
        return None

    def doSelection(self, objID):
        if self.liveSelections == True:
            if objID in self.objIDs:
                print("doSelection: {}".format(objID))
                symbIdx = self.objIDs.index(objID)
                self.set_sentence(self.sentence.text + self.getSymb(symbIdx) )

    def set_sentence(self, text):
        '''set/update the text to show in the instruction screen'''
        if type(text) is list:
            text = "\n".join(text)
        self.sentence.begin_update()
        self.sentence.text=text
        self.sentence.end_update()

    def set_grid(self, symbols=None, objIDs=None, bgFraction=.3, sentence="What you type goes here"):
        '''set/update the grid of symbols to be selected from'''
        winw, winh=window.get_size()
        # tell noisetag which objIDs we are using
        if symbols is not None:
            self.symbols = symbols
        else:
            symbols = self.symbols
        # Number of non-None symbols
        nsymb      = sum([sum([(s is not None) for s in x ]) for x in symbols])

        if objIDs is not None:
            self.objIDs = objIDs
        else:
            if self.objIDs is None:
                self.objIDs = list(range(1,nsymb+1))
            objIDs = self.objIDs
        # get size of the matrix
        gridheight  = len(symbols) + 1 # extra row for sentence
        gridwidth = max([len(s) for s in symbols])
        ngrid      = gridwidth * gridheight

        self.noisetag.setActiveObjIDs(self.objIDs)

        # add a background sprite with the right color
        self.objects=[None]*nsymb
        self.labels=[None]*nsymb
        self.batch = pyglet.graphics.Batch()
        self.background = pyglet.graphics.OrderedGroup(0)
        self.foreground = pyglet.graphics.OrderedGroup(1)

        # now create the display objects
        w=winw/gridwidth # cell-width
        bgoffsetx = w*bgFraction
        h=winh/gridheight # cell-height
        bgoffsety = h*bgFraction
        idx=-1
        for i in range(len(symbols)): # rows
            y = (gridheight-1-i-1)/gridheight*winh # top-edge cell
            for j in range(len(symbols[i])): # cols
                # skip unused positions
                if symbols[i][j] is None: continue
                idx = idx+1
                x = j/gridwidth*winw # left-edge cell
                # create a 1x1 white image for this grid cell
                img = pyglet.image.SolidColorImagePattern(color=(255, 255, 255, 255)).create_image(2, 2)
                # convert to a sprite (for fast re-draw) and store in objects list
                # and add to the drawing batch (as background)
                self.objects[idx]=pyglet.sprite.Sprite(img, x=x+bgoffsetx, y=y+bgoffsety,
                                                       batch=self.batch, group=self.background)
                # re-scale (on GPU) to the size of this grid cell
                self.objects[idx].update(scale_x=int(w-bgoffsetx*2)/img.width,
                                         scale_y=int(h-bgoffsety*2)/img.height)
                # add the foreground label for this cell, and add to drawing batch
                self.labels[idx]=pyglet.text.Label(symbols[i][j], font_size=32, x=x+w/2, y=y+h/2,
                                                   color=(255, 255, 255, 255),
                                                   anchor_x='center', anchor_y='center',
                                                   batch=self.batch, group=self.foreground)

        # add opto-sensor block
        img = pyglet.image.SolidColorImagePattern(color=(255, 255, 255, 255)).create_image(1, 1)
        self.opto_sprite=pyglet.sprite.Sprite(img, x=0, y=winh*.9,
                                              batch=self.batch, group=self.background)
        self.opto_sprite.update(scale_x=int(winw*.1), scale_y=int(winh*.1))
        self.opto_sprite.visible=False

        # add the sentence box
        y = (gridheight-1)/gridheight*winh # top-edge cell
        x = winw*.2 # left-edge cell
        self.sentence=pyglet.text.Label(sentence, font_size=32, x=x, y=y+h/2,
                                        color=(255, 255, 255, 255),
                                        anchor_x='left', anchor_y='center',
                                        batch=self.batch, group=self.foreground)

    def is_done(self):
        if self.isDone:
            self.noisetag.modeChange('idle')
        return self.isDone

    # mapping from bci-stimulus-states to display color
    state2color={0:(10, 10, 10),   # off=grey
                 1:(255, 255, 255), # on=white
                 2:(0, 255, 0),    # cue=green
                 3:(0, 0, 255)}    # feedback=blue
    def draw(self, t):
        """draw the letter-grid with given stimulus state for each object.
        Note: To maximise timing accuracy we send the info on the grid-stimulus state
        at the start of the *next* frame, as this happens as soon as possible after
        the screen 'flip'. """
        if not self.isRunning:
            self.isRunning=True
        self.framestart=self.noisetag.getTimeStamp()
        winflip = window.lastfliptime
        if winflip > self.framestart or winflip < self.frameend:
            print("Error: frameend={} winflip={} framestart={}".format(self.frameend,winflip,self.framestart))
        self.nframe = self.nframe+1
        if self.sendEvents:
            self.noisetag.sendStimulusState(timestamp=winflip)#self.frameend)#window.lastfliptime)

        # get the current stimulus state to show
        try:
            self.noisetag.updateStimulusState()
            stimulus_state, target_state, objIDs, sendEvents=self.noisetag.getStimulusState()
        except StopIteration:
            self.isDone=True
            return

        if self.waitKey:
            global last_key_press
            if self.window.last_key_press:
                self.key_press = self.window.last_key_press
                #self.noisetag.reset()
                self.isDone = True
                self.window.last_key_press = None

        # turn all off if no stim-state
        if stimulus_state is None:
            stimulus_state = [0]*len(self.objects)

        # do the stimulus callback if wanted
        if self.stimulus_callback is not None:
            self.stimulus_callback(stimulus_state, target_state)

        # draw the white background onto the surface
        if self.clearScreen:
            window.clear()
        # update the state
        # TODO[]: iterate over objectIDs and match with those from the
        #         stimulus state!
        for idx in range(min(len(self.objects), len(stimulus_state))):
            # set background color based on the stimulus state (if set)
            try:
                self.objects[idx].color=self.state2color[stimulus_state[idx]]
                self.labels[idx].color=(255,255,255,255) # reset labels
            except KeyError:
                pass


        # show live-feedback (if wanted)
        if self.liveFeedback:
            # get prediction info if any
            predMessage=self.noisetag.getLastPrediction()
            if predMessage and predMessage.Yest in objIDs and predMessage.Perr < self.feedbackThreshold:
                predidx=objIDs.index(predMessage.Yest) # convert from objID -> objects index
                prederr=predMessage.Perr
                # BODGE: manually mix in the feedback color as blue tint on the label
                fbcol=list(self.labels[predidx].color)
                fbcol[0]=fbcol[0]*.4 
                if fbcol[0]>1: fbcol[0]=int(fbcol[0])
                fbcol[1]=fbcol[1]*.4; 
                if fbcol[1]>1: fbcol[1]=int(fbcol[1])
                fbcol[2]=fbcol[2]*.4+255*(1-prederr)*.6; 
                if fbcol[2]>1: fbcol[2]=int(fbcol[2])
                self.labels[predidx].color=fbcol

        # disp opto-sensor if targetState is set
        if self.optosensor :
            if self.opto_sprite is not None:
                self.opto_sprite.visible=False  # default to opto-off
            if target_state is not None and target_state in (0, 1):
                print("*" if target_state==1 else '.', end='', flush=True)
                if self.opto_sprite is not None:
                    self.opto_sprite.visible=True
                    self.opto_sprite.color = (0, 0, 0) if target_state==0 else (255, 255, 255)

        # do the draw
        self.batch.draw()
        self.frameend=self.noisetag.getTimeStamp()

        # frame flip time logging info
        if self.LOGLEVEL > 0 and self.noisetag.isConnected():
            opto = target_state if target_state is not None else 0
            logstr="FrameIdx:%d FlipTime:%d FlipLB:%d FlipUB:%d Opto:%d"%(nframe, self.framestart, self.framestart, self.frameend, opto)
            self.noisetag.log(logstr)




#-----------------------------------------------------------------
#-----------------------------------------------------------------
#-----------------------------------------------------------------
#---------------------------------------------------------
from enum import IntEnum
class ExptScreenManager(Screen):
    '''class to manage a whole application, with main menu, checks, etc.'''

    class ExptPhases(IntEnum):
        ''' enumeration for the different phases of an experiment/BCI application '''
        MainMenu=0
        Connecting=1
        SignalQuality=2
        CalInstruct=3
        Calibration=4
        CalResults=5
        CuedPredInstruct=6
        CuedPrediction=7
        PredInstruct=8
        Prediction=9
        Closing=10
        Quit=100
        Welcome=99
        Minimize=101
        Settings=105
        FrameRateCheck=200
        Reset=110

    welcomeInstruct="Welcome to the mindaffectBCI\n\nkey to continue"
    calibrationInstruct="Calibration\n\nThe next stage is CALIBRATION\nlook at the indicated green target\n\nkey to continue"
    cuedpredictionInstruct="Prediction\n\nThe next stage is CUED PREDICTION\nLook at the green cued letter\n\nLive BCI feedback in blue\n\nkey to continue"
    predictionInstruct="Prediction\n\nThe next stage is free PREDICTION\nLook at the letter you want to select\nLive BCI feedback in blue\n\nkey to continue"
    closingInstruct="Closing\nThankyou\n\nPress to exit"
    resetInstruct="Reset\n\nThe decoder model has been reset.\nYou will need to run calibration again to use the BCI\n\nkey to continue"

    main_menu ="Welcome to the mindaffectBCI" +"\n"+ \
               "\n"+ \
               "Press the number of the option you want:" +"\n"+ \
               "\n"+ \
               "0) Electrode Quality" +"\n"+ \
               "1) Calibration" +"\n"+ \
               "2) Cued Prediction" +"\n"+ \
               "3) Free Typing" +"\n"+ \
               "Q) Quit" + "\n\n\n" + \
               "f) frame-rate-check" + "\n" + \
               "s) settings\n" + \
               "r) reset calibration model"
               
    menu_keys = {pyglet.window.key._0:ExptPhases.SignalQuality,
                 pyglet.window.key._1:ExptPhases.CalInstruct,
                 pyglet.window.key._2:ExptPhases.CuedPredInstruct,
                 pyglet.window.key._3:ExptPhases.PredInstruct,
                 pyglet.window.key.F:ExptPhases.FrameRateCheck,
                 pyglet.window.key.S:ExptPhases.Settings,
                 pyglet.window.key.R:ExptPhases.Reset,
                 pyglet.window.key.Q:ExptPhases.Quit}

    def __init__(self, window, noisetag, symbols, nCal:int=1, nPred:int=1, framesperbit:int=None, fullscreen_stimulus:bool=True, selectionThreshold:float=.1, optosensor:bool=True):
        self.window = window
        self.noisetag = noisetag
        self.symbols = symbols
        self.menu = MenuScreen(window, self.main_menu, self.menu_keys.keys())
        self.instruct = InstructionScreen(window, '', duration = 50000)
        self.connecting = ConnectingScreen(window, noisetag)
        self.query  =  QueryDialogScreen(window, 'Query Test:')
        self.electquality = ElectrodequalityScreen(window, noisetag)
        self.results = ResultsScreen(window, noisetag)
        self.selectionGrid = SelectionGridScreen(window, symbols, noisetag, optosensor=optosensor)
        self.stage = self.ExptPhases.Connecting
        self.next_stage = self.ExptPhases.Connecting
        self.nCal = nCal
        self.nPred = nPred
        self.framesperbit = framesperbit
        self.fullscreen_stimulus = fullscreen_stimulus
        self.selectionThreshold = selectionThreshold
        self.screen = None
        self.transitionNextPhase()

    def draw(self, t):
        if self.screen is None:
            return
        self.screen.draw(t)
        if self.screen.is_done():
            self.transitionNextPhase()

    def is_done(self):
        return self.screen is None

    def transitionNextPhase(self):
        print("stage transition")

        # move to the next stage
        if self.next_stage is not None:
            self.stage = self.next_stage
            self.next_stage = None
        else: # assume it's from the menu
            self.stage = self.menu_keys.get(self.menu.key_press,self.ExptPhases.MainMenu)
            self.next_stage = None

        if self.stage==self.ExptPhases.MainMenu: # main menu
            if self.fullscreen_stimulus==True :
                self.window.set_fullscreen(fullscreen=False)

            print("main menu")
            self.menu.reset()
            self.screen = self.menu
            self.noisetag.modeChange('idle')
            self.next_stage = None
            

        elif self.stage==self.ExptPhases.Welcome: # welcome instruct
            print("welcome instruct")
            self.instruct.set_text(self.welcomeInstruct)
            self.instruct.reset()
            self.screen = self.instruct
            self.next_stage = self.ExptPhases.Connecting

        elif self.stage==self.ExptPhases.Reset: # reset the decoder
            print("reset")
            self.instruct.set_text(self.resetInstruct)
            self.instruct.reset()
            self.screen = self.instruct
            self.noisetag.modeChange("reset")
            self.next_stage = self.ExptPhases.MainMenu

        elif self.stage==self.ExptPhases.Connecting: # connecting instruct
            print("connecting screen")
            self.connecting.reset()
            self.screen = self.connecting
            self.next_stage = self.ExptPhases.MainMenu

        elif self.stage==self.ExptPhases.SignalQuality: # electrode quality
            print("signal quality")
            self.electquality.reset()
            self.screen=self.electquality
            self.next_stage = self.ExptPhases.MainMenu

        elif self.stage==self.ExptPhases.CalInstruct: # calibration instruct
            print("Calibration instruct")
            if self.fullscreen_stimulus==True :
                self.window.set_fullscreen(fullscreen=True)
            self.instruct.set_text(self.calibrationInstruct)
            self.instruct.reset()
            self.screen=self.instruct
            self.next_stage = self.ExptPhases.Calibration

        elif self.stage==self.ExptPhases.Calibration: # calibration
            print("calibration")
            self.selectionGrid.noisetag.startCalibration(nTrials=self.nCal, numframes=4.2/isi, waitduration=1, framesperbit=self.framesperbit)
            self.selectionGrid.reset()
            self.selectionGrid.liveFeedback=False
            self.selectionGrid.set_sentence('Calibration: look at the green cue.')
            self.screen = self.selectionGrid
            self.next_stage = self.ExptPhases.CalResults

        elif self.stage==self.ExptPhases.CalResults: # Calibration Results
            print("Calibration Results")
            if self.fullscreen_stimulus==True :
                self.window.set_fullscreen(fullscreen=True)
            self.results.reset()
            self.screen=self.results
            self.next_stage = self.ExptPhases.MainMenu

        elif self.stage==self.ExptPhases.CuedPredInstruct: # pred instruct
            print("cued pred instruct")
            if self.fullscreen_stimulus==True :
                self.window.set_fullscreen(fullscreen=True)
            self.instruct.set_text(self.cuedpredictionInstruct)
            self.instruct.reset()
            self.screen=self.instruct
            self.next_stage = self.ExptPhases.CuedPrediction

        elif self.stage==self.ExptPhases.CuedPrediction: # pred
            print("cued prediction")
            self.selectionGrid.noisetag.startPrediction(nTrials=self.nPred, numframes=10/isi, cuedprediction=True, waitduration=1, framesperbit=self.framesperbit, selectionThreshold=self.selectionThreshold)
            self.selectionGrid.reset()
            self.selectionGrid.liveFeedback=True
            self.selectionGrid.setliveSelections(True)
            self.selectionGrid.set_sentence('CuedPrediction: look at the green cue.\n')
            self.screen = self.selectionGrid
            self.next_stage = self.ExptPhases.MainMenu

        elif self.stage==self.ExptPhases.PredInstruct: # pred instruct
            print("pred instruct")
            if self.fullscreen_stimulus==True :
                self.window.set_fullscreen(fullscreen=True)
            self.instruct.set_text(self.predictionInstruct)
            self.instruct.reset()
            self.screen=self.instruct
            self.next_stage = self.ExptPhases.Prediction

        elif self.stage==self.ExptPhases.Prediction: # pred
            print("prediction")
            self.selectionGrid.noisetag.startPrediction(nTrials=self.nPred, numframes=10/isi, cuedprediction=False, waitduration=1, framesperbit=self.framesperbit, selectionThreshold=self.selectionThreshold)
            self.selectionGrid.reset()
            self.selectionGrid.liveFeedback=True
            self.selectionGrid.set_sentence('')
            self.selectionGrid.setliveSelections(True)
            self.screen = self.selectionGrid
            self.next_stage = self.ExptPhases.MainMenu

        elif self.stage==self.ExptPhases.Closing: # closing instruct
            print("closing instruct")
            self.instruct.set_text(self.closingInstruct)
            self.instruct.reset()
            self.screen=self.instruct
            self.next_stage = self.ExptPhases.Quit

        elif self.stage==self.ExptPhases.Minimize: # closing instruct
            print("minimize")
            self.window.minimize()
            self.next_stage = self.ExptPhases.MainMenu

        elif self.stage==None: # testing stages..
            #print("flicker with selection")
            #self.selectionGrid.noisetag.startFlickerWithSelection(numframes=10/isi)
            print("single trial")
            self.selectionGrid.set_grid([[None, 'up', None],
                                         ['left', 'fire', 'right']])
            self.selectionGrid.noisetag.startSingleTrial(numframes=10/isi)
            # N.B. ensure decoder is in prediction mode!
            self.selectionGrid.noisetag.modeChange('Prediction.static')
            self.selectionGrid.reset()
            self.screen = self.selectionGrid

        elif self.stage==self.ExptPhases.FrameRateCheck: # frame-rate-check
            print("frame-rate-check")
            #from mindaffectBCI.examples.presentation.framerate_check import FrameRateTestScreen
            self.screen=FrameRateTestScreen(self.window,waitKey=True)
            self.next_stage = self.ExptPhases.MainMenu

        elif self.stage==self.ExptPhases.Settings: # config settings
            print("settings")
            self.screen = SettingsScreen(self.window, self)
            self.next_stage = self.ExptPhases.MainMenu

        else: # end
            print('quit')
            self.screen=None


#------------------------------------------------------------------------
# Initialization: display, utopia-connection
# use noisetag object as time-stamp provider
def getTimeStamp():
    if 'nt' in globals():
        global nt
        return nt.getTimeStamp()
    else: # fall back if not connected to utopia client
        import time
        return (int(time.perf_counter()*1000) % (1<<31))

import types
from mindaffectBCI.noisetag import sumstats
flipstats=sumstats(60)
fliplogtime=0
def timedflip(self):
    '''pseudo method type which records the timestamp for window flips'''
    global flipstats, fliplogtime
    type(self).flip(self)
    olft=self.lastfliptime
    self.lastfliptime=getTimeStamp()
    if flipstats is not None:
        flipstats.addpoint(self.lastfliptime-olft)
        #if self.lastfliptime > fliplogtime:
        #    fliplogtime=fliplogtime+5000
        #    print("\nFlipTimes:"+str(flipstats))
        #    print("Hist:\n"+flipstats.hist())

def on_key_press(symbols, modifiers):
    '''main key-press handler, which stores the last key in a global variable'''
    global window
    window.last_key_press=symbols

def on_text(text):
    global window
    window.last_text = text

def initPyglet(fullscreen=False):
    '''intialize the pyglet window, keyhandler'''
    global window
    # set up the window
    if fullscreen:
        # N.B. accurate video timing only guaranteed with fullscreen
        # N.B. over-sampling seems to introduce frame lagging on windows+Intell
        config = pyglet.gl.Config(double_buffer=True) #double_buffer=False,sample_buffers=1, samples=4)
        window = pyglet.window.Window(fullscreen=True, vsync=True, config=config)
    else:
        config = pyglet.gl.Config(double_buffer=True)#,sample_buffers=1, samples=4)
        window = pyglet.window.Window(width=1024, height=768, vsync=True, resizable=True, config=config)

    # setup a key press handler, just store key-press in global variable
    window.push_handlers(on_key_press, on_text)
    window.last_key_press=None
    window.last_text=None

    global nframe; nframe=0
    # override window's flip method to record the exact *time* the
    # flip happended
    window.flip = types.MethodType(timedflip, window)
    window.lastfliptime=getTimeStamp()
    global fliplogtime; fliplogtime=window.lastfliptime

    # minimize on tab away, when in fullscreen mode:
    @window.event
    def on_deactivate():
        # TODO []: stop minimise when switch to/from full-screen mode
        if fullscreen:
            window.minimize()

    # TODO[]: handle resize events correctly.
    return window

def draw(dt):
    '''main window draw function, which redirects to the screen stack'''
    global ss, nframe
    nframe=nframe+1
    ss.draw(dt)
    # check for termination
    if ss.is_done():
        print('app exit')
        pyglet.app.exit()
    #print('.', end='', flush=True)

def load_symbols(fn):
    """load a screen layout from a text file

    Args:
        fn (str): file name to load from

    Returns:
        symbols [list of lists of str]: list of list of the symbols strings
    """
    symbols = []

    # look relative to the py-file dir if can't find
    import os.path
    if not os.path.isfile(fn):
        pydir = os.path.dirname(os.path.abspath(__file__))
        fn = os.path.join(pydir, fn)

    with open(fn,'r') as f:
        for line in f:
            # skip comment lines
            if line.startswith('#'): continue
            # delim is ,
            line = line.split(',')
            # strip whitespace
            line = [ l.strip() for l in line if l is not None ]
            # add
            symbols.append(line)

    return symbols

def run(symbols=None, ncal:int=10, npred:int=10, stimfile=None, selectionThreshold:float=None,
        framesperbit:int=1, optosensor:bool=True, fullscreen:bool=False, windowed:bool=None, fullscreen_stimulus:bool=False, simple_calibration=False, host=None):
    """ run the selection Matrix with default settings

    Args:
        nCal (int, optional): number of calibration trials. Defaults to 10.
        nPred (int, optional): number of prediction trials at a time. Defaults to 10.
        simple_calibration (bool, optional): flag if we show only a single target during calibration, Defaults to False.
        stimFile ([type], optional): the stimulus file to use for the codes. Defaults to None.
        framesperbit (int, optional): number of video frames per stimulus codebit. Defaults to 1.
        fullscreen (bool, optional): flag if should runn full-screen. Defaults to False.
        fullscreen_stimulus (bool, optional): flag if should run the stimulus (i.e. flicker) in fullscreen mode. Defaults to True.
    """
    global nt, ss
    # N.B. init the noise-tag first, so asks for the IP
    if stimfile is None:
        stimfile = 'mgold_61_6521_psk_60hz.txt'
    if fullscreen is None and windowed is not None:
        fullscreen = not windowed
    if windowed:
        fullscreen_stimulus = False
    nt=Noisetag(stimFile=stimfile,clientid='Presentation:selectionMatrix')
    if host is not None and not host in ('','-'):
        nt.connect(host, queryifhostnotfound=False)

    # init the graphics system
    initPyglet(fullscreen=fullscreen)

    # the logical arrangement of the display matrix
    if symbols is None:
        symbols=[['a', 'b', 'c', 'd', 'e'],
                 ['f', 'g', 'h', 'i', 'j'],
                 ['k', 'l', 'm', 'n', 'o'],
                 ['p', 'q', 'r', 's', 't'],
                 ['u', 'v', 'w', 'x', 'y']]

    elif isinstance(symbols,str):
        # load the layout from a file
        symbols = load_symbols(symbols)

    # make the screen manager object which manages the app state
    ss = ExptScreenManager(window, nt, symbols, nCal=ncal, nPred=npred, framesperbit=framesperbit, fullscreen_stimulus=fullscreen_stimulus, selectionThreshold=selectionThreshold, optosensor=optosensor)

    # set per-frame callback to the draw function
    if drawrate > 0:
        # slow down for debugging
        pyglet.clock.schedule_interval(draw, drawrate)
    else:
        # call the draw method as fast as possible, i.e. at video frame rate!
        pyglet.clock.schedule(draw)
    # mainloop
    pyglet.app.run()


if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('ncal',type=int, help='number calibration trials', nargs='?', default=10)
    parser.add_argument('npred',type=int, help='number prediction trials', nargs='?', default=10)
    parser.add_argument('--host',type=str, help='address (IP) of the utopia-hub', default=None)
    parser.add_argument('--stimfile',type=str, help='stimulus file to use', default=None)
    parser.add_argument('--framesperbit',type=int, help='number of video frames per stimulus bit', default=1)
    #parser.add_argument('--fullscreen',action='store_true',help='run in fullscreen mode')
    parser.add_argument('--windowed',action='store_true',help='run in fullscreen mode')
    parser.add_argument('--selectionThreshold',type=float,help='target error threshold for selection to occur',default=.1)
    #parser.add_argument('--simple_calibration',action='store_true',help='flag to only show a single target during calibration')
    #parser.add_option('-m','--matrix',action='store',dest='symbols',help='file with the set of symbols to display',default=None)
    args = parser.parse_args()

    run(**vars(args))




import os
import numpy as np
import re
from mindaffectBCI.utopiaclient import StimulusEvent, DataPacket, ModeChange, NewTarget, Selection
from mindaffectBCI.decoder.utils import unwrap

# named reg-exp to parse the different messages types log lines
recievedts_re = re.compile(r'.*rts:(?P<rts>[-0-9]*)\W')
serverts_re = re.compile(r'.*sts:(?P<sts>[-0-9]*)\W')
clientts_re = re.compile(r'^.*\Wts:(?P<ts>[-0-9]*)\W')
clientip_re = re.compile(r'.*<-\W(?P<ip>[0-9.:]*)$')
stimevent_re = re.compile(r'^.*\Wts:(?P<ts>[-0-9]*)\W*v\[(?P<shape>[0-9x]*)\]:(?P<stimstate>.*) <-.*$')
datapacket_re = re.compile(r'^.*\Wts:(?P<ts>[-0-9]*)\W*v\[(?P<shape>[0-9x]*)\]:(?P<samples>.*) <-.*$')
modechange_re = re.compile(r'^.*\Wts:(?P<ts>[-0-9]*)\W.*mode:(?P<newmode>.*) <-.*$')

def read_StimulusEvent(line:str):
    ''' read a stimulus event message from a text-file save version '''
    # named reg-ex to extract the bits we need
    res = stimevent_re.match(line)
    if res is None:
        return None
    ts = int(res['ts'])
    # parse sample into numpy array
    shape = np.fromstring(res['shape'].replace('x',','),sep=',',dtype=int)
    shape = shape[::-1] # N.B. python order, fastest last..
    stiminfo = res['stimstate'].replace('{','').replace('}',',')
    stiminfo = np.fromstring(stiminfo, sep=',', dtype=int)
    objIDs = stiminfo[0::2]
    stimstate = stiminfo[1::2]
    #print("SE ts:{} objIDs:{} state:{}".format(ts,objIDs,stimstate))
    return StimulusEvent(ts,objIDs,stimstate)
    
def read_DataPacket(line:str ):
    # named reg-ex to extract the bits we need
    res = datapacket_re.match(line)
    if res is None:
        return None
    ts = int(res['ts'])
    # parse sample into numpy array
    shape = np.fromstring(res['shape'].replace('x',','),sep=',',dtype=int)
    shape = shape[::-1] # N.B. python order, fastest last..
    samples = np.fromstring(res['samples'].replace(']','').replace('[',''),sep=',',dtype=np.float32)
    samples = samples[:np.prod(shape)].reshape(shape) # guard too many samples?
    return DataPacket(ts,samples)
    
def read_ModeChange(line:str):
    res = modechange_re.match(line)
    if res is None:
        return None
    ts = int(res['ts'])
    newmode = res['newmode']
    return ModeChange(ts,newmode)

def read_NewTarget(line:str):
    ts = read_clientts(line)
    return NewTarget(ts)

def read_Selection(line:str):
    # TODO[]: read the actual selection info
    match = read_clientts(line)
    return Selection(match,-1)

def read_clientts(line:str):
    match = clientts_re.match(line)
    match = int(match['ts']) if match is not None else None
    return match

def read_serverts(line:str):
    match = serverts_re.match(line)
    match = int(match['sts']) if match is not None else None
    return match

def read_recievedts(line:str):
    match = recievedts_re.match(line)
    match = int(match['rts']) if match is not None else None
    return match

def read_clientip(line:str):
    ip = clientip_re.match(line)
    ip = ip['ip'] if ip is not None else None
    return ip

def read_mindaffectBCI_message(line:str):
    """Read a mindaffectBCI message from a line of text

    Args:
        line (str): A line containing a mindaffectBCI message

    Returns:
        [type]: The decoded message as a message class.
    """    
    if StimulusEvent.msgName in line:
        msg = read_StimulusEvent(line)
    elif DataPacket.msgName in line:
        msg = read_DataPacket(line)
    elif ModeChange.msgName in line:
        msg = read_ModeChange(line)
    elif NewTarget.msgName in line:
        msg = read_NewTarget(line)
    elif Selection.msgName in line:
        msg = read_Selection(line)
    else:
        msg = None
    # add the server time-stamp
    if msg :
        msg.sts = read_serverts(line) # server time-samp
        msg.rts = read_recievedts(line) # recieved time-stamp
        msg.clientip = read_clientip(line) # client ip-address
    return msg

def datapackets2array(msgs, sample2timestamp='lower_bound_tracker', timestamp_ch=None):
    """Convert a set of datapacket messages to a 2-d numpy array (with timestamp channel)

    Args:
        msgs ([type]): list of DataPacket messages
        sample2timestamp (str, optional): filtering function, to filter the data-packet time-stamps using the increasing sample counts. Defaults to 'lower_bound_tracker'.
        timestamp_ch (int, optional): If set, channel which contains the timestamp information. Defaults to None.

    Returns:
        X( (t,d) np.ndarray): the extracted samples in a numpy array
    """
    data=[]
    from mindaffectBCI.decoder.UtopiaDataInterface import timestamp_interpolation, linear_trend_tracker

    # extract the time-stamp channel and map to server time-stamps
    # and insert as the packet time-stamp
    if timestamp_ch is not None and timestamp_ch < len(msgs[0].samples[0]):
        # get the client-timestamp, server-timestamp pairs
        x = np.array([msg.samples[-1,timestamp_ch] for msg in msgs]) #  from: client timestamp, from ts-channel
        y = np.array([msg.rts for msg in msgs]) # to: server timestamp
        ab = robust_timestamp_regression(x,y)
        # now rewrite the client time-stamps
        for m in msgs:
            m.rawtimestamp = m.timestamp # save the old one
            m.timestamp = m.samples[-1,timestamp_ch]*ab[0] + ab[1]
            # wrap into the desired bit-size if wanted
            if hasattr(m,'unwrapped_timestamp'):
                m.unwrapped_timestamp = m.timestamp
                # BODGE: fixed wrapping size!
                m.timestamp = m.timestamp % (1<<24)
            
    tsfilt = timestamp_interpolation(sample2timestamp=sample2timestamp)
    for m in msgs:
        samples = m.samples
        ts   = m.timestamp
        # TODO[]: look at using the time-stamps non-filtered?
        samples_ts = tsfilt.transform(ts,len(samples))
        samples = np.append(samples, samples_ts[:,np.newaxis], -1).astype(samples.dtype)
        data.append(samples)
        #last_ts = ts
        
    # convert data into single np array
    data = np. concatenate(data,0)
    return data    


def robust_timestamp_regression(x,y):
    # Warning: strip and -1....
    invalidts = np.logical_or(x==-1,y==-1)
    x = x[~invalidts]
    x = unwrap(x)
    y = y[~invalidts]
    y = unwrap(y)
    # add constant feature  for the intercept
    x = np.append(x[:,np.newaxis],np.ones((x.shape[0],1)),1)
    # LS  solve
    # TODO[X]: use a proper weighted least squares robust estimator, also for on-line
    sqrt_w = np.ones(y.shape,dtype=x.dtype)
    for i in range(4):
        ab,res,_,_ = np.linalg.lstsq(x*sqrt_w[:,np.newaxis],y*sqrt_w,rcond=-1)
        y_est = x[:,0]*ab[0] + ab[1]
        err = y - y_est # server > true, clip positive errors
        scale = np.mean(np.abs(err[sqrt_w>0]))

        sqrt_w[:]=1
        clipIdx = err > 3*scale
        sqrt_w[clipIdx]=0
        #print("{} overestimates".format(np.sum(clipIdx)))
        clipIdx = err < -3*scale
        sqrt_w[clipIdx]=0
        #print("{} underestimates".format(np.sum(clipIdx)))
    return ab


def rewrite_timestamps2servertimestamps(msgs):
    ''' rewrite message time-stamps  to best-fit server time-stamps '''
    # get the client-timestamp, server-timestamp pairs
    x = np.array([msg.timestamp for msg in msgs]) #  from: client timestamp
    y = np.array([msg.sts for msg in msgs]) # to: server timestamp
    ab = robust_timestamp_regression(x,y)

    #print("ab={}".format(ab))
    # now rewrite the client time-stamps
    for m in msgs:
        m.rawtimestamp = m.timestamp
        m.timestamp = m.rawtimestamp*ab[0] + ab[1]
    return msgs

    
def read_mindaffectBCI_messages( fn:str, regress:bool=False ):
    ''' read the data from a text-file save version into a list of messages,
        WARNING: this reads the  messages as raw, and does *not* try to time-stamp clocks
                 w.r.t.  the message source.  To compare messages between clients you will
                 need to do this manually! '''
    fn = os.path.expanduser(fn)
    with open(fn,'r') as file:
        msgs=[]
        for line in file:
            msg = read_mindaffectBCI_message(line)
            if msg is not None:
                msgs.append(msg)

    # TODO [X]: intelligent time-stamp re-writer taking account of the client-ip
    if regress is None:
        # do nothing, leave client + server time-stamps in place
        pass
    elif regress==False: 
        # just use the server ts, as they have been put on the same time-line by the hub
        for m in msgs:
            m.origal_timestamp = m.timestamp
            m.timestamp = m.sts
    else:
        clientips = [ m.clientip for m in msgs ]
        for client in set(clientips):
            clientmsgs = [ c for c in msgs if c.clientip == client ]
            # only regress datapacket messages
            #if not any(isinstance(m,DataPacket) for m in clientmsgs):
            print('rewrite for client ip: {}'.format(client))
            rewrite_timestamps2servertimestamps(clientmsgs)
            #else:
            #    for m in msgs:
            #        m.timestamp = m.sts
        
    return msgs

def read_mindaffectBCI_data_messages( fn:str, regress=False, timestamp_wrap_size=(1<<24), **kwargs ):
    ''' read the data from a text-file save version into a dataarray and message list '''
    rawmsgs = read_mindaffectBCI_messages(fn, regress)
    # split into datapacket messages and others
    data=[]
    msgs=[]
    for m in rawmsgs:
        #  WARNING BODGE: fit time-stamp in 24bits for float32 ring buffer
        #  Note: this leads to wrap-arroung in (1<<24)/1000/3600 = 4.6 hours
        #        but that shouldn't matter.....
        if timestamp_wrap_size is not None:
            m.unwrapped_timestamp = m.timestamp
            m.timestamp = m.timestamp % timestamp_wrap_size

        if isinstance(m,DataPacket):
            data.append(m)
        else:
            msgs.append(m)

    # convert the data messages into a single numpy array,
    # with (interpolated) time-stamps in the final 'channel'
    data = datapackets2array(data, **kwargs)
    
    return (data,msgs)


def testcase(fn=None):
    ''' testcase, load reference datafile '''
    if fn is None:
        fn = 'mindaffectBCI.txt'

    print("read messages from {}".format(fn))
    msgs = read_mindaffectBCI_messages(fn)
    for msg in msgs[:100]:
        print("{}".format(msg))

    print("read data messages")
    data,msgs = read_mindaffectBCI_data_messages(fn)
    print("Data({})={}".format(data.shape,data))
    for m in msgs[:100]:
        print("{}".format(m))

    
if __name__=="__main__":
    # default to last log file if not given
    import glob
    import os
    fileregexp = '../../../logs/mindaffectBCI*.txt'
    #fileregexp = '../../../../utopia/java/utopia2ft/UtopiaMessages*.log'
    files = glob.glob(os.path.join(os.path.dirname(os.path.abspath(__file__)),os.path.expanduser(fileregexp))) # * means all if need specific format then *.csv
    fn = max(files, key=os.path.getctime)
    #if len(sys.argv) > 0:
    #    fn = sys.argv[1]
    testcase(fn)



#!/usr/bin/python3
# Copyright (c) 2019 MindAffect B.V. 
#  Author: Jason Farquhar <jason@mindaffect.nl>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from mindaffectBCI.utopiaclient import *

def doAction(msgs):
    for msg in msgs:
        # skip non-selection messages
        if not msg.msgID==Selection.msgID :
            continue
        try:
            # get the function to execute for selections we are responsible for
            selectionAction = objectID2Actions[msg.objID]
            # run the action function
            selectionAction(msg.objID)
        except KeyError:
            # This is fine, just means it's a selection we are not responsible for
            Pass

# the set of actions to perform
def forward(objID):
    print('move forward')
def backward(objID):
    print('move backward')
def left(objID):
    print('move left')
def right(objID):
    print('move right')

# map from objectIDs to the function to execute
objectID2Actions = { 64:forward, 65:backward, 66:left, 67:right }

client = UtopiaClient()
client.autoconnect()
client.sendMessage(Subscribe(client.getTimeStamp(),"S"))
while True:
    newmsgs=client.getNewMessages()
    doAction(newmsgs)
